# pylint: disable=redefined-builtin,no-member

from __future__ import annotations

import json
import re
from typing import Any, Optional, Type, TypeVar
from uuid import uuid4

from fideslang.validation import validate_fides_key
from sqlalchemy import Column, DateTime, String
from sqlalchemy.ext.declarative import declarative_base, declared_attr
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import Query, Session
from sqlalchemy.sql import func
from sqlalchemy.sql.expression import BinaryExpression, BooleanClauseList
from sqlalchemy_utils import JSONType

from fides.api.common_exceptions import KeyOrNameAlreadyExists, KeyValidationError
from fides.api.util.custom_json_encoder import CustomJSONEncoder, _custom_decoder
from fides.api.util.text import to_snake_case

T = TypeVar("T", bound="OrmWrappedFidesBase")
ALLOWED_CHARS_PATTERN = r"[A-Za-z0-9\-_]"
DISALLOWED_CHARS_PATTERN = (
    r"[^A-Za-z0-9\-_]"  # anything that's _not_ matched by ALLOWED_CHARS_PATTERN above
)
ALLOWED_CHARS = re.compile(ALLOWED_CHARS_PATTERN)
DISALLOWED_CHARS = re.compile(DISALLOWED_CHARS_PATTERN)


class JSONTypeOverride(JSONType):  # pylint: disable=W0223
    """
    Temporary override of JSONType with workaround to fix the return type.

    Using a StringEncryptedType with JSONType as the underlying type returns a string instead of a dictionary:
    '{"key": "value"}' instead of {"key": "value"}

    https://github.com/kvesteri/sqlalchemy-utils/issues/532
    """

    def process_bind_param(self, value: MutableDict, _: Any | None) -> str | None:
        """
        Overrides JSONType.process_bind_param to return json.dumps(value) instead of just the value.
        """
        if value is not None:
            return json.dumps(value, cls=CustomJSONEncoder)
        return value

    def process_result_value(self, value: str, _: Any | None) -> dict[str, Any] | None:
        """
        Overrides JSONType.process_result_value to return json.loads(value) instead of just the value.
        """
        if value is not None:
            return json.loads(value, object_hook=_custom_decoder)
        return value


def get_key_from_data(data: dict[str, Any], cls_name: str) -> str:
    """Extracts key from data, validates.

    If no key, uses a snake-cased name. Will be used as the URL onupdate
    applicable classes.
    """
    supplied_key = data.get("key")
    if supplied_key and not isinstance(supplied_key, str):  # Mypy check
        raise KeyValidationError("Supplied key must be a string")

    key = validate_fides_key(supplied_key) if supplied_key else None
    if key is None:
        name = data.get("name")
        if name is None:
            raise KeyValidationError(f"{cls_name} requires a name.")
        key = validate_fides_key(to_snake_case(name))
    return key


class FidesBase:
    """
    A generic base class to be used for all DB models, automatically adding the
    following fields to every table:
    - id: a UUID
    - created_at: the datetime that the record was created
    - updated_at: the datetime that the record was last updated
    """

    def generate_uuid(self) -> str:
        """
        Generates a uuid with a prefix based on the tablename to be used as the
        record's ID value
        """
        try:
            # `self` in this context is an instance of
            # sqlalchemy.dialects.postgresql.psycopg2.PGExecutionContext_psycopg2
            prefix = f"{self.current_column.table.name[:3]}_"  # type: ignore
        except AttributeError:
            # If the table name is unavailable for any reason, we don't
            # need to use it
            prefix = ""
        uuid = str(uuid4())
        return f"{prefix}{uuid}"

    @declared_attr
    def __tablename__(self) -> str:
        """The name of this model's table in the DB"""
        return self.__name__.lower()  # type: ignore #pylint: disable=E1101

    @declared_attr
    def id_field_path(self) -> str:
        """
        The database path to any model's ID field, for use with ForeignKey
        specifications in the event we have overridden a model's
        __tablename__ attribute
        """
        return f"{self.__tablename__}.id"

    @classmethod
    def sanitize_key(cls, proposed_key: str) -> str:
        """
        Sanitize the key by removing invalid characters.

        Invalid characters are based on the allowed characters defined in this module.
        Note that this differs slightly from the allowed characters in `sanitize_fides_key`
        from `core/utils.py`; specifically, `.` are disallowed (converted to `_`) here, while
        they are allowed in `sanitize_fides_key`. This is because `.` must be allowed in
        `fides_key`s, but they are not allowed in `key`s.
        """
        return DISALLOWED_CHARS.sub("_", proposed_key)

    id = Column(String(255), primary_key=True, index=True, default=generate_uuid)
    created_at = Column(DateTime(timezone=True), server_default=func.now())
    updated_at = Column(
        DateTime(timezone=True), server_default=func.now(), onupdate=func.now()
    )


class OrmWrappedFidesBase(FidesBase):
    """A wrapper class for our base model.

    This allows us to include the base model in type checking for our return types.
    This abstraction includes the following methods:
    - get(id): return the record at that particular ID
    - all(): return all records in that table
    - filter(conditions): return all records that satisfy filter conditions
    - create(data): create a record with provided data
    - update_from_class(conditions, values): update all records that match the filter
        conditions with the data inside values, called from the class
    - update: update the record related to the object calling this method
    - delete_with_class(id): delete the record at the provided ID, called from the class
    - delete: delete the record related to the object calling this method
    - delete_all: delete all records in this table
    - save: update the record related to the object calling this method with the current
        data stored on the object
    """

    __mapper_args__ = {"confirm_deleted_rows": False}
    # This line hides sql alchemy warnings of the form:
    # Sqlalchemy/orm/persistence.py:1461: SAWarning: DELETE statement on table
    # 'storageconfig' expected to delete 1 row(s); 0 were matched. Please set
    # confirm_deleted_rows=False within the mapper configuration to prevent this
    # warning.

    @classmethod
    def get_optional_field_names(cls) -> list[str]:
        """Returns the names of all nullable fields on the wrapped model."""
        return [
            field.name
            for field in list(cls.__table__.columns)  # type: ignore # pylint: disable=E1101
            if field.nullable
        ]

    @classmethod
    def get(cls: Type[T], db: Session, *, object_id: Any) -> T | None:
        """Fetch a database record via a table ID."""
        return db.query(cls).get(object_id)

    @classmethod
    def get_by(
        cls: Type[T],
        db: Session,
        *,
        field: str,
        value: str | int,
    ) -> T | None:
        """Fetch a database record via a dynamic key, supplied at call time."""
        kwargs = {field: value}
        return db.query(cls).filter_by(**kwargs).first()

    @classmethod
    def query(cls: Type[T], db: Session) -> Query:
        """Create a blank query for the class."""
        return db.query(cls)

    @classmethod
    def all(cls: Type[T], db: Session) -> list[T]:
        """Fetch all database records in table."""
        return db.query(cls).all()

    @classmethod
    def filter(
        cls: Type[T],
        db: Session,
        *,
        conditions: BinaryExpression | BooleanClauseList,
    ) -> Query:
        """Fetch multiple models from a database table."""
        return db.query(cls).filter(conditions)

    @classmethod
    def create(
        cls: Type[T],
        db: Session,
        *,
        data: dict[str, Any],
        check_name: bool = True,
    ) -> T:
        """Create a new row in the database."""
        # Build properly formatted key for applicable classes
        if hasattr(cls, "key"):
            data["key"] = get_key_from_data(data, cls.__name__)
            if db.query(cls).filter_by(key=data["key"]).first():
                raise KeyOrNameAlreadyExists(
                    f"Key {data['key']} already exists in {cls.__name__}. Keys will be snake-cased names if not provided. "
                    f"If you are seeing this error without providing a key, please provide a key or a different name."
                    ""
                )

        if check_name and hasattr(cls, "name"):
            data["name"] = data.get("name")
            if db.query(cls).filter_by(name=data["name"]).first():
                raise KeyOrNameAlreadyExists(
                    f"Name {data['name']} already exists in {cls.__name__}."
                )

        # Create
        db_obj = cls(**data)  # type: ignore
        return cls.persist_obj(db, db_obj)

    @classmethod
    def get_by_key_or_id(
        cls: Type[T], db: Session, *, data: dict[str, Any]
    ) -> T | None:
        """Retrieves db object by id, if provided, otherwise attempts by key."""
        db_obj = None
        if data.get("id") is not None:
            db_obj = cls.get(db=db, object_id=data["id"])
        elif data.get("key") is not None:
            db_obj = cls.get_by(db=db, field="key", value=data["key"])
        return db_obj

    @classmethod
    def create_or_update(
        cls: Type[T],
        db: Session,
        *,
        data: dict[str, Any],
        check_name: Optional[bool] = True,
    ) -> T:
        """Create an object, or update the existing version.

        There's an edge case where `data["id"]` and `data["key"]` can point attempt
        different records, in which case this method will attempt to update the
        fetched record with the key of another, and a `KeyOrNameAlreadyExists`
        error will be thrown. If neither `key`, nor `id` are passed in, leave
        `db_obj` as None and assume we are creating a new object.
        """
        db_obj = cls.get_by_key_or_id(db=db, data=data)
        if db_obj:
            db_obj.update(db=db, data=data)  # type: ignore
        else:
            db_obj = cls.create(db=db, data=data, check_name=check_name)  # type: ignore
        return db_obj

    @classmethod
    def get_or_create(
        cls: Type[T], db: Session, *, data: dict[str, Any]
    ) -> tuple[bool, T | None]:
        """Fetch an object, or create it if none is found."""
        db_obj = db.query(cls).filter_by(**data).first()
        created = False
        if not db_obj:
            db_obj = cls.create(db=db, data=data)  # type: ignore
            created = True
        return created, db_obj

    @classmethod
    def update_with_class(
        cls: Type[T], db: Session, *, conditions: Any, values: dict[str, Any]
    ) -> int:
        """Update all objects within a filter at database level."""
        return db.query(cls).filter(conditions).update(values=values)

    @classmethod
    def delete_with_class(cls: Type[T], db: Session, *, id: str) -> T | None:
        """Delete an existing row from the database from the object's class."""
        obj = db.query(cls).get(id)
        if obj is None:
            return None
        db.delete(obj)
        db.commit()
        return obj

    @classmethod
    def delete_all(cls: Type[T], db: Session) -> int:
        """Delete all rows in this table."""
        deleted_count = db.query(cls).delete()
        return deleted_count

    def refresh_from_db(self, db: Session) -> FidesBase | None:
        """Returns a current version of this object from the database."""
        return db.query(self.__class__).get(self.id)

    def update(self, db: Session, *, data: dict[str, Any]) -> FidesBase:
        """Update specific row with supplied values."""
        # Set self.key where applicable
        if hasattr(self, "key") and "key" in data:
            data["key"] = get_key_from_data(data, self.__class__.__name__)

        # Update dataset with values in data
        for key, val in data.items():
            setattr(self, key, val)

        return self.save(db=db)

    def delete(self, db: Session) -> FidesBase | None:
        """Delete an existing row in the database from an existing object in memory."""
        db.delete(self)
        db.commit()
        return self

    def save(self, db: Session) -> FidesBase:
        """Save the current object over an existing row in the database."""
        if hasattr(self, "key"):
            key = getattr(self, "key")

            is_valid = key is not None and (len(ALLOWED_CHARS.sub("", key)) == 0)

            if not is_valid:
                raise KeyValidationError(
                    f"Key '{key}' on {self.__class__.__name__} is invalid."
                )

        return OrmWrappedFidesBase.persist_obj(db, self)

    @classmethod
    def persist_obj(cls: Type[T], db: Session, resource: T) -> T:
        """Method to be run after 'create' or 'save' to write the resource to the db
        Can be overridden on subclasses to not commit immediately when creating
        creating/updating.
        """
        db.add(resource)
        db.commit()
        db.refresh(resource)
        return resource


Base = declarative_base(cls=OrmWrappedFidesBase)



"""S3-specific storage client implementation."""

from __future__ import annotations

from typing import Any, Optional

from fideslang.validation import AnyHttpUrlString
from loguru import logger

from fides.api.schemas.storage.storage import AWSAuthMethod
from fides.api.service.storage.s3 import create_presigned_url_for_s3
from fides.api.service.storage.streaming.base_storage_client import BaseStorageClient
from fides.api.util.aws_util import get_s3_client


class S3StorageClient(BaseStorageClient):
    """S3-specific storage client implementation.

    Handles S3-specific URI construction, transport parameters, and presigned URL
    generation for the smart-open storage client.
    """

    def __init__(self, auth_method: str, storage_secrets: dict[str, Any]):
        """Initialize the storage client with secrets.

        Args:
            storage_secrets: Provider-specific storage credentials and configuration using string keys
                           (e.g., "aws_access_key_id", "region_name") from format_secrets()
        """
        super().__init__(storage_secrets)
        self.storage_secrets: dict[str, Any] = storage_secrets
        self.auth_method = auth_method

    def build_uri(self, bucket: str, key: str) -> str:
        """Build S3 URI for the given bucket and key.

        Args:
            bucket: S3 bucket name
            key: Object key/path

        Returns:
            S3 URI string for smart-open
        """
        # Handle custom endpoint (e.g., MinIO) - endpoint_url is not in StorageSecrets enum
        # For now, we'll use standard S3 URI since we only have enum keys
        return f"s3://{bucket}/{key}"

    def get_transport_params(self) -> dict[str, Any]:
        """Get S3-specific transport parameters for smart-open.
        Type annotation: get_s3_client returns a boto3 S3 client object, not a Session
        This is what smart-open expects for the "client" transport parameter

        Returns:
            Dictionary of S3 transport parameters for smart-open
        """
        params: dict[str, Any] = {}

        # Create S3 client for smart-open
        try:
            # Determine auth method based on available credentials
            if self.auth_method == AWSAuthMethod.AUTOMATIC.value:

                # For automatic authentication, check if region is available
                if not self.storage_secrets.get("region_name", None):
                    logger.warning(
                        "No region specified in storage secrets for automatic authentication"
                        "This may cause credential issues - consider setting a default region"
                    )

            # Extract assume_role_arn if present
            assume_role_arn = None
            if (
                "assume_role_arn" in self.storage_secrets
                and self.storage_secrets["assume_role_arn"]
            ):
                assume_role_arn = self.storage_secrets["assume_role_arn"]
                logger.debug(f"Using assume role ARN: {assume_role_arn}")

            # Create S3 client using existing utility
            # get_s3_client returns a boto3 S3 client, not a Session
            s3_client: Any = None
            try:
                s3_client = get_s3_client(
                    self.auth_method, self.storage_secrets, assume_role_arn  # type: ignore
                )
                logger.debug("Successfully created S3 client")
            except Exception as e:
                # For automatic authentication, try to provide more helpful error messages
                if self.auth_method == AWSAuthMethod.AUTOMATIC.value:
                    logger.error(
                        f"Failed to create S3 client with automatic authentication: {e}. "
                        "This usually means AWS credentials are not available in the environment"
                        "Please ensure AWS credentials are configured via environment variables, IAM roles, or AWS profiles"
                    )
                    raise ValueError(
                        f"Automatic AWS authentication failed: {e}. Please check your AWS credential configuration."
                    )
                raise

            params["client"] = s3_client

        except Exception as e:
            logger.error(f"Failed to create S3 client for smart-open: {e}")
            raise

        # Include credentials at top level for compatibility
        # Note: When using an S3 client, these credential parameters are not needed
        # and will be ignored by smart-open, causing warnings
        # Only include them if no S3 client is provided (fallback scenario)
        if not params.get("client"):
            for key, transport_key in [
                ("aws_access_key_id", "access_key"),
                ("aws_secret_access_key", "secret_key"),
                ("region_name", "region"),
                ("assume_role_arn", "assume_role_arn"),
            ]:
                if key in self.storage_secrets and self.storage_secrets[key]:
                    params[transport_key] = self.storage_secrets[key]

        return params

    def generate_presigned_url(
        self, bucket: str, key: str, ttl_seconds: Optional[int] = None
    ) -> AnyHttpUrlString:
        """Generate an S3 presigned URL for the object.

        Args:
            bucket: S3 bucket name
            key: Object key/path
            ttl_seconds: Time to live in seconds

        Returns:
            S3 presigned URL for the object

        Raises:
            Exception: If presigned URL generation fails
        """
        try:
            # Extract assume_role_arn if present for role assumption
            assume_role_arn = None
            if "assume_role_arn" in self.storage_secrets:
                assume_role_arn = self.storage_secrets["assume_role_arn"]

            # get_s3_client returns a boto3 S3 client, not a Session
            s3_client: Any = get_s3_client(
                self.auth_method, self.storage_secrets, assume_role_arn  # type: ignore
            )
            return create_presigned_url_for_s3(s3_client, bucket, key, ttl_seconds)
        except Exception as e:
            logger.error(f"Failed to generate S3 presigned URL: {e}")
            raise



"""remove client id constraint

Revision ID: 68cb26f3492d
Revises: 956d21f13def
Create Date: 2024-02-01 18:07:16.757838

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "68cb26f3492d"
down_revision = "956d21f13def"
branch_labels = None
depends_on = None


def upgrade():
    op.alter_column("policy", "client_id", existing_type=sa.VARCHAR(), nullable=True)


def downgrade():
    op.alter_column("policy", "client_id", existing_type=sa.VARCHAR(), nullable=False)



"""add_manual_task_to_connectiontype_enum

Revision ID: aadfe83c5644
Revises: 6a76a1fa4f3f
Create Date: 2025-06-19 18:55:08.131278

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "aadfe83c5644"
down_revision = "6a76a1fa4f3f"
branch_labels = None
depends_on = None


def upgrade():
    # Add manual_task to ConnectionType enum
    op.execute("alter type connectiontype rename to connectiontype_old")
    op.execute(
        "create type connectiontype as enum('attentive_email', 'bigquery', 'datahub', 'dynamodb', 'fides', 'generic_consent_email', 'generic_erasure_email', 'dynamic_erasure_email', 'google_cloud_sql_mysql', 'google_cloud_sql_postgres', 'https', 'manual', 'manual_webhook', 'manual_task', 'mariadb', 'mongodb', 'mssql', 'mysql', 'okta', 'postgres', 'rds_mysql', 'rds_postgres', 'redshift', 's3', 'saas', 'scylla', 'snowflake', 'sovrn', 'timescale', 'website')"
    )
    op.execute(
        (
            "alter table connectionconfig alter column connection_type type connectiontype using "
            "connection_type::text::connectiontype"
        )
    )
    op.execute("drop type connectiontype_old")


def downgrade():
    # Remove manual_task from ConnectionType enum
    op.execute("alter type connectiontype rename to connectiontype_old")
    op.execute(
        "create type connectiontype as enum('attentive_email', 'bigquery', 'datahub', 'dynamodb', 'fides', 'generic_consent_email', 'generic_erasure_email', 'dynamic_erasure_email', 'google_cloud_sql_mysql', 'google_cloud_sql_postgres', 'https', 'manual', 'manual_webhook', 'mariadb', 'mongodb', 'mssql', 'mysql', 'okta', 'postgres', 'rds_mysql', 'rds_postgres', 'redshift', 's3', 'saas', 'scylla', 'snowflake', 'sovrn', 'timescale', 'website')"
    )
    op.execute(
        (
            "alter table connectionconfig alter column connection_type type connectiontype using "
            "connection_type::text::connectiontype"
        )
    )
    op.execute("drop type connectiontype_old")



"""
Execution context for collecting log messages during privacy request processing.

Uses Python's contextvars for thread-safe, execution-scoped message collection.
"""

import contextvars
from contextlib import contextmanager
from typing import Generator, List, Optional

from loguru import logger

# Context variable for current execution messages
_execution_messages: contextvars.ContextVar[Optional[List[str]]] = (
    contextvars.ContextVar("execution_messages")
)


@contextmanager
def collect_execution_log_messages() -> Generator[List[str], None, None]:
    """
    Context manager for collecting execution log messages.

    Usage:
        with collect_execution_log_messages() as messages:
            # ... do work that calls add_execution_log_message()
            pass
        # messages now contains all collected messages

    Returns:
        List[str]: Messages collected during execution
    """
    messages: List[str] = []
    token = _execution_messages.set(messages)
    try:
        yield messages
    finally:
        _execution_messages.reset(token)


def add_execution_log_message(message: str) -> None:
    """Add a message to the current execution context"""
    try:
        messages = _execution_messages.get()
        if messages is not None and message.strip():
            messages.append(message.strip())
    except LookupError:
        logger.error(
            f"Attempted to call add_execution_log_message outside of execution context, unable to add message: {message}"
        )


def get_execution_log_messages() -> List[str]:
    """Get messages from current execution context (for testing/debugging)"""
    try:
        messages = _execution_messages.get()
        return messages.copy() if messages is not None else []
    except LookupError:
        return []



from typing import Union

from fides.api.schemas.base_class import NoValidationSchema
from fides.api.schemas.storage.storage import StorageSecretsGCS, StorageSecretsS3


class StorageSecretsS3Docs(StorageSecretsS3, NoValidationSchema):
    """The secrets required to connect to S3, for documentation"""


class StorageSecretsGCSDocs(StorageSecretsGCS, NoValidationSchema):
    """The secrets required to connect to Google Cloud Storage, for documentation"""


possible_storage_secrets = Union[StorageSecretsS3Docs, StorageSecretsGCSDocs]



import datetime
from typing import Any, List, Literal, Optional, Union

from fastapi_pagination import Page, Params
from fastapi_pagination.ext.async_sqlalchemy import paginate as async_paginate
from fideslang.validation import FidesKey
from sqlalchemy import or_
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.future import select
from sqlalchemy.sql.selectable import Select

from fides.api.db.crud import list_resource
from fides.api.models.sql_models import (  # type: ignore[attr-defined]
    PrivacyDeclaration,
    System,
)
from fides.api.schemas.filter_params import FilterParams
from fides.api.util.filter_utils import apply_filters_to_query


class SystemService:
    """
    Service for system management. Currently only used for listing systems.
    """

    def __init__(self, db: AsyncSession):
        self.db = db

    async def get_systems(
        self,
        *,
        search: Optional[str] = None,
        data_uses: Optional[List[FidesKey]] = None,
        data_categories: Optional[List[FidesKey]] = None,
        data_subjects: Optional[List[FidesKey]] = None,
        size: Optional[int] = None,
        page: Optional[int] = None,
        show_deleted: bool = False,
        sort_by: Optional[List[Literal["name"]]] = None,
        sort_asc: Optional[bool] = True,
        **kwargs: Any,
    ) -> Union[List[System], Page[System]]:
        """
        Query systems with flexible filtering and optional pagination.

        This function provides the core logic for listing systems with support for:
        - Basic filtering (search, data uses, categories, subjects)
        - Pagination
        - Vendor deletion filtering
        - Extension points for additional filtering

        Args:
            db: Database session
            size: Page size for pagination (if None, no pagination)
            page: Page number for pagination
            show_deleted: Whether to include vendor-deleted systems
            filter_params: Standard filter parameters (search, taxonomy filters)

        Returns:
            Either a list of all systems or a paginated page of systems
        """

        filter_params = None
        if any([search, data_uses, data_categories, data_subjects]):
            filter_params = FilterParams(
                search=search,
                data_uses=data_uses,
                data_categories=data_categories,
                data_subjects=data_subjects,
            )

        # If no parameters are provided, return all systems for backward compatibility
        if not any([size, page, filter_params]):
            return await list_resource(System, self.db)

        query = self._build_systems_query(
            filter_params=filter_params,
            show_deleted=show_deleted,
            **kwargs,
        )

        # Ensure distinct results
        if sort_by:
            for sort_column_name in sort_by:
                if sort_column_name == "name":
                    name_order = System.name.asc() if sort_asc else System.name.desc()
                    # For name sorting, we need name first in ORDER BY, then id for uniqueness
                    query = query.distinct(System.name, System.id).order_by(
                        name_order, System.id
                    )
        else:
            # Default to ascending name sort when no sorting parameters are provided
            name_order = System.name.asc()
            query = query.distinct(System.name, System.id).order_by(
                name_order, System.id
            )

        # Return paginated or non-paginated results
        if size or page:
            pagination_params = Params(page=page or 1, size=size or 50)
            return await async_paginate(self.db, query, pagination_params)

        result = await self.db.execute(query)
        return list(result.scalars().all())

    def _build_systems_query(
        self,
        *,
        filter_params: Optional[FilterParams] = None,
        show_deleted: bool = False,
        **kwargs: Any,
    ) -> Select:
        """
        Build the base query for systems.
        """

        query = select(System)

        # Determine if we need to join with PrivacyDeclaration
        needs_privacy_declaration = filter_params and any(
            [
                filter_params.data_uses,
                filter_params.data_categories,
                filter_params.data_subjects,
            ]
        )

        if needs_privacy_declaration:
            query = query.outerjoin(
                PrivacyDeclaration, System.id == PrivacyDeclaration.system_id
            )

        # Filter out vendor deleted systems unless explicitly asked for
        if not show_deleted:
            query = query.filter(
                or_(
                    System.vendor_deleted_date.is_(None),
                    System.vendor_deleted_date >= datetime.datetime.now(),
                )
            )

        # Apply standard filters if provided
        if filter_params:
            query = apply_filters_to_query(
                query=query,
                filter_params=filter_params,
                search_model=System,
                taxonomy_model=(
                    PrivacyDeclaration if needs_privacy_declaration else None
                ),
            )

        return query



from typing import List, Set

from sqlalchemy import Column, String
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import Session

from fides.api.db.base_class import Base


class PrivacyRequestRedactionPattern(Base):
    """
    Stores one regex pattern per row for masking dataset, collection, and field names
    in DSR package reports.
    """

    @declared_attr
    def __tablename__(self) -> str:
        return "privacy_request_redaction_pattern"

    # One pattern per row; unique to prevent duplicates
    pattern = Column(String, nullable=False, unique=True)

    @classmethod
    def get_patterns(cls, db: Session) -> List[str]:
        """
        Get the current list of masking patterns ordered by creation time.

        Returns:
            List of regex pattern strings, or None if no patterns are configured
        """
        rows = db.query(cls).order_by(cls.created_at.asc()).all()
        if not rows:
            return []
        return [row.pattern for row in rows]

    @classmethod
    def replace_patterns(cls, db: Session, patterns: List[str]) -> List[str]:
        """
        Replace the set of stored patterns with the provided list using set reconciliation.

        - Adds patterns that are not present
        - Removes patterns that are no longer desired
        - Returns the resulting canonical list in deterministic order
        """
        # Normalize: trim whitespace, remove empties and duplicates
        desired: Set[str] = {p.strip() for p in patterns if p and p.strip()}

        # Fetch existing patterns from DB as a set
        existing: Set[str] = set(p for (p,) in db.query(cls.pattern).all())

        to_add = desired - existing
        to_remove = existing - desired

        if to_remove:
            db.query(cls).filter(cls.pattern.in_(list(to_remove))).delete(
                synchronize_session=False
            )

        if to_add:
            db.bulk_save_objects([cls(pattern=p) for p in to_add])

        db.commit()

        return sorted(desired)



import re
from typing import List

from pydantic import BaseModel, Field, field_validator


class PrivacyRequestRedactionPatternsRequest(BaseModel):
    """Request schema for updating privacy request redaction patterns."""

    patterns: List[str] = Field(
        description="List of regex patterns used to redact dataset, collection, and field names in privacy request package reports",
        max_length=100,  # Limit number of patterns
    )

    @field_validator("patterns")
    @classmethod
    def validate_patterns(cls, patterns: List[str]) -> List[str]:
        """Validate regex patterns with ReDoS protection via Pydantic's default rust-regex engine."""
        if not patterns:
            return patterns

        validated_patterns = []
        for i, pattern in enumerate(patterns):
            if not isinstance(pattern, str):
                raise ValueError(f"Pattern at index {i} must be a string")

            pattern = pattern.strip()
            if not pattern:
                raise ValueError(f"Pattern at index {i} cannot be empty")

            # Reasonable length limit for regex patterns
            if len(pattern) > 500:
                raise ValueError(
                    f"Pattern at index {i} is too long (max 500 characters)"
                )

            # Pydantic's rust-regex engine provides ReDoS protection
            try:
                re.compile(pattern, re.IGNORECASE)
            except re.error as e:
                raise ValueError(f"Invalid regex pattern at index {i}: {e}")

            validated_patterns.append(pattern)

        return validated_patterns


class PrivacyRequestRedactionPatternsResponse(BaseModel):
    """Response schema for privacy request redaction patterns."""

    patterns: List[str] = Field(
        description="List of regex patterns used to redact dataset, collection, and field names in privacy request package reports"
    )

    model_config = {"from_attributes": True}



"""add event audit table

Revision ID: 5fa78b1f324d
Revises: fd7571bef683
Create Date: 2025-09-09 17:51:15.715976

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "5fa78b1f324d"
down_revision = "fd7571bef683"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "event_audit",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("event_type", sa.String(), nullable=False),
        sa.Column("user_id", sa.String(), nullable=True),
        sa.Column("resource_type", sa.String(), nullable=True),
        sa.Column("resource_identifier", sa.String(), nullable=True),
        sa.Column("description", sa.Text(), nullable=True),
        sa.Column(
            "event_details", postgresql.JSONB(astext_type=sa.Text()), nullable=True
        ),
        sa.Column(
            "status",
            sa.Enum("succeeded", "failed", name="eventauditstatus"),
            nullable=False,
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_event_audit_event_type"), "event_audit", ["event_type"], unique=False
    )
    op.create_index(op.f("ix_event_audit_id"), "event_audit", ["id"], unique=False)
    op.create_index(
        op.f("ix_event_audit_resource_identifier"),
        "event_audit",
        ["resource_identifier"],
        unique=False,
    )
    op.create_index(
        op.f("ix_event_audit_resource_type"),
        "event_audit",
        ["resource_type"],
        unique=False,
    )
    op.create_index(
        op.f("ix_event_audit_user_id"), "event_audit", ["user_id"], unique=False
    )
    op.create_index(
        op.f("ix_event_audit_status"), "event_audit", ["status"], unique=False
    )


def downgrade():
    op.drop_index(op.f("ix_event_audit_status"), table_name="event_audit")
    op.drop_index(op.f("ix_event_audit_user_id"), table_name="event_audit")
    op.drop_index(op.f("ix_event_audit_resource_type"), table_name="event_audit")
    op.drop_index(op.f("ix_event_audit_resource_identifier"), table_name="event_audit")
    op.drop_index(op.f("ix_event_audit_id"), table_name="event_audit")
    op.drop_index(op.f("ix_event_audit_event_type"), table_name="event_audit")
    op.drop_table("event_audit")



"""Update Request Tasks

Revision ID: 90502bcda282
Revises: 3baf42d251a6
Create Date: 2025-08-05 16:31:45.335384

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "90502bcda282"
down_revision = "3baf42d251a6"
branch_labels = None
depends_on = None

async_task_type = sa.Enum("manual", "polling", "callback", name="asynctasktype")


def upgrade():
    async_task_type.create(op.get_bind(), checkfirst=True)
    op.add_column(
        "requesttask",
        sa.Column(
            "async_type",
            async_task_type,
            nullable=True,
        ),
    )


def downgrade():
    op.drop_column("requesttask", "async_type")
    async_task_type.drop(op.get_bind(), checkfirst=True)



from typing import Any, Dict, List, Optional

from loguru import logger
from requests import head

from fides.api.common_exceptions import ConnectionException
from fides.api.graph.execution import ExecutionNode
from fides.api.models.connectionconfig import ConnectionTestStatus
from fides.api.models.policy import Policy
from fides.api.models.privacy_request import PrivacyRequest, RequestTask
from fides.api.service.connectors.base_connector import BaseConnector
from fides.api.service.connectors.query_configs.query_config import QueryConfig
from fides.api.util.collection_util import Row


class WebsiteConnector(BaseConnector):
    """
    Website connector, used currently for Website 'monitoring' - this is class is used to test basic connections to the website

    NOTE: No DSR processing is supported for Website connectors.
    """

    def dsr_supported(self) -> bool:
        return False

    def create_client(self) -> Any:  # type: ignore
        """Returns a client for the website"""

    def query_config(self, node: ExecutionNode) -> QueryConfig[Any]:
        """DSR execution not supported for Website connectors"""
        raise NotImplementedError()

    def test_connection(self) -> Optional[ConnectionTestStatus]:
        """
        Validates the connection to the website by executing a `HEAD` request against the provided URL.

        TODO: can we perform a better validation by pinging the website from the web monitor proxy?
        TODO: can we validate credentials somehow when we support them?
        """
        website_url = self.configuration.secrets.get("url")
        logger.info(
            "Starting test connection to connector '{}', at URL '{}'",
            self.configuration.key,
            website_url,
        )
        try:
            response = head(self.configuration.secrets["url"])
        except Exception as error:
            raise ConnectionException(str(error))

        if response.status_code >= 400:
            raise ConnectionException(
                f"HEAD request to '{website_url}' resulted in error status code: '{response.status_code}'"
            )

        logger.info("Connection to '{}' succeeded", website_url)

        return ConnectionTestStatus.succeeded

    def retrieve_data(
        self,
        node: ExecutionNode,
        policy: Policy,
        privacy_request: PrivacyRequest,
        request_task: RequestTask,
        input_data: Dict[str, List[Any]],
    ) -> List[Row]:
        """DSR execution not supported for website connector"""
        return []

    def mask_data(
        self,
        node: ExecutionNode,
        policy: Policy,
        privacy_request: PrivacyRequest,
        request_task: RequestTask,
        rows: List[Row],
        input_data: Optional[Dict[str, List[Any]]] = None,
    ) -> int:
        """DSR execution not supported for website connector"""
        return 0

    def close(self) -> None:
        """Close any held resources"""
        # no held resources for website connector



"""Test that DSR 3.0 is now the default scheduler with backward compatibility for existing DSR 2.0 requests."""

from fides.api.models.privacy_request import PrivacyRequest
from fides.api.models.privacy_request.request_task import RequestTask
from fides.api.schemas.policy import ActionType
from fides.api.task.scheduler_utils import use_dsr_3_0_scheduler


class TestDSR3DefaultScheduler:
    """Test that DSR 3.0 is now the default scheduler."""

    def test_dsr_3_0_is_default_for_new_requests(self, db, policy):
        """Test that new privacy requests use DSR 3.0 by default."""
        # Create a new privacy request without any previous processing
        privacy_request = PrivacyRequest.create(
            db=db,
            data={
                "policy_id": policy.id,
                "status": "pending",
            },
        )

        # The scheduler should default to DSR 3.0 for new requests
        result = use_dsr_3_0_scheduler(privacy_request, ActionType.access)
        assert result is True

        privacy_request.delete(db=db)

    def test_existing_dsr_2_0_requests_can_continue(self, db, policy, cache):
        """Test that existing DSR 2.0 requests can continue with DSR 2.0."""
        # Create a privacy request and simulate it having previous DSR 2.0 results
        privacy_request = PrivacyRequest.create(
            db=db,
            data={
                "policy_id": policy.id,
                "status": "in_processing",
            },
        )

        # Simulate having previous DSR 2.0 results by storing data in cache
        cache_key = (
            f"{privacy_request.id}__access_request__test_dataset:test_collection"
        )
        cache.set_encoded_object(cache_key, {"test": "data"})

        # The scheduler should allow existing DSR 2.0 requests to continue
        result = use_dsr_3_0_scheduler(privacy_request, ActionType.access)
        assert result is False  # Should continue using DSR 2.0

        privacy_request.delete(db=db)
        cache.delete(cache_key)

    def test_new_requests_without_cache_use_dsr_3_0(self, db, policy):
        """Test that new requests without existing cache data use DSR 3.0."""
        # Create a privacy request without any cache data
        privacy_request = PrivacyRequest.create(
            db=db,
            data={
                "policy_id": policy.id,
                "status": "pending",
            },
        )

        # Should use DSR 3.0 for new requests
        result = use_dsr_3_0_scheduler(privacy_request, ActionType.access)
        assert result is True

        privacy_request.delete(db=db)

    def test_cache_exists_but_tasks_exist_uses_dsr_3_0(self, db, policy, cache):
        """
        Test that when cache exists but tasks also exist, scheduler uses DSR 3.0.

        This verifies the existing_tasks_count condition in the scheduler decision.
        The scheduler only allows DSR 2.0 when cache exists AND no tasks exist.
        """
        # Create a privacy request
        privacy_request = PrivacyRequest.create(
            db=db,
            data={
                "policy_id": policy.id,
                "status": "in_processing",
            },
        )

        # Simulate having previous DSR 2.0 results by storing data in cache
        cache_key = (
            f"{privacy_request.id}__access_request__test_dataset:test_collection"
        )
        cache.set_encoded_object(cache_key, {"test": "data"})

        # Create an existing task to simulate partial processing with DSR 3.0
        RequestTask.create(
            db,
            data={
                "action_type": ActionType.access,
                "status": "pending",
                "privacy_request_id": privacy_request.id,
                "collection_address": "test_dataset:test_collection",
                "dataset_name": "test_dataset",
                "collection_name": "test_collection",
                "upstream_tasks": [],
                "downstream_tasks": [],
                "all_descendant_tasks": [],
            },
        )

        # Even though cache exists, because tasks exist, should use DSR 3.0
        result = use_dsr_3_0_scheduler(privacy_request, ActionType.access)
        assert result is True  # Should use DSR 3.0 because tasks exist



import json
from typing import Dict, List

from loguru import logger
from redis.lock import Lock
from sqlalchemy import text
from sqlalchemy.orm import Session

from fides.api.db.session import get_db_session
from fides.api.tasks.scheduled.scheduler import scheduler
from fides.api.util.lock import redis_lock
from fides.config import CONFIG

"""
This utility is used to detect indices or constraints that were deferred as part
of a standard Fides/Alembic migration.

The standard approach for creating indices or constraints on large tables is a blocking operation,
meaning reads and write can't run while the indices and constraints are being created.
A non-blocking approach is to use the CONCURRENTLY keyword for index creation then adding
the constraints with ADD CONSTRAINT USING INDEX.

However these approaches cannot run in a transaction, which is what Fides/Alembic uses to
apply migrations. The approach here is to run these commands after the application has started
up and commit the changes immediately, outside of a transaction.

The commands that need to run are denoted in the `TABLE_OBJECT_MAP`. Indices that are used
to create constraints also include the `constraint_name` value since either the presence of
the index or the constraint can be used to determine if an index has executed.
When using the ADD CONSTRAINT USING INDEX syntax, the index specified in the command is
automatically renamed to match the name of the constraint being added.
This means that after the constraint is created, the index will have the same name as the constraint.
"""

POST_UPGRADE_INDEX_CREATION = "post_upgrade_index_creation"

TABLE_OBJECT_MAP: Dict[str, List[Dict[str, str]]] = {
    "currentprivacypreferencev2": [
        {
            "name": "ix_currentprivacypreferencev2_email_property_id",
            "statement": "CREATE UNIQUE INDEX CONCURRENTLY ix_currentprivacypreferencev2_email_property_id ON currentprivacypreferencev2 (email, property_id)",
            "type": "index",
            "constraint_name": "last_saved_for_email_per_property_id",
        },
        {
            "name": "ix_currentprivacypreferencev2_external_id_property_id",
            "statement": "CREATE UNIQUE INDEX CONCURRENTLY ix_currentprivacypreferencev2_external_id_property_id ON currentprivacypreferencev2 (external_id, property_id)",
            "type": "index",
            "constraint_name": "last_saved_for_external_id_per_property_id",
        },
        {
            "name": "ix_currentprivacypreferencev2_fides_user_device_property_id",
            "statement": "CREATE UNIQUE INDEX CONCURRENTLY ix_currentprivacypreferencev2_fides_user_device_property_id ON currentprivacypreferencev2 (fides_user_device, property_id)",
            "type": "index",
            "constraint_name": "last_saved_for_fides_user_device_per_property_id",
        },
        {
            "name": "ix_currentprivacypreferencev2_phone_number_property_id",
            "statement": "CREATE UNIQUE INDEX CONCURRENTLY ix_currentprivacypreferencev2_phone_number_property_id ON currentprivacypreferencev2 (phone_number, property_id)",
            "type": "index",
            "constraint_name": "last_saved_for_phone_number_per_property_id",
        },
        {
            "name": "ix_currentprivacypreferencev2_hashed_external_id",
            "statement": "CREATE INDEX CONCURRENTLY ix_currentprivacypreferencev2_hashed_external_id ON currentprivacypreferencev2 (hashed_external_id)",
            "type": "index",
        },
        {
            "name": "last_saved_for_email_per_property_id",
            "statement": "ALTER TABLE currentprivacypreferencev2 ADD CONSTRAINT last_saved_for_email_per_property_id UNIQUE USING INDEX ix_currentprivacypreferencev2_email_property_id",
            "type": "constraint",
        },
        {
            "name": "last_saved_for_external_id_per_property_id",
            "statement": "ALTER TABLE currentprivacypreferencev2 ADD CONSTRAINT last_saved_for_external_id_per_property_id UNIQUE USING INDEX ix_currentprivacypreferencev2_external_id_property_id",
            "type": "constraint",
        },
        {
            "name": "last_saved_for_fides_user_device_per_property_id",
            "statement": "ALTER TABLE currentprivacypreferencev2 ADD CONSTRAINT last_saved_for_fides_user_device_per_property_id UNIQUE USING INDEX ix_currentprivacypreferencev2_fides_user_device_property_id",
            "type": "constraint",
        },
        {
            "name": "last_saved_for_phone_number_per_property_id",
            "statement": "ALTER TABLE currentprivacypreferencev2 ADD CONSTRAINT last_saved_for_phone_number_per_property_id UNIQUE USING INDEX ix_currentprivacypreferencev2_phone_number_property_id",
            "type": "constraint",
        },
    ],
    "privacypreferencehistory": [
        {
            "name": "ix_privacypreferencehistory_hashed_external_id",
            "statement": "CREATE INDEX CONCURRENTLY ix_privacypreferencehistory_hashed_external_id ON privacypreferencehistory (hashed_external_id)",
            "type": "index",
        },
    ],
    "servednoticehistory": [
        {
            "name": "ix_servednoticehistory_hashed_external_id",
            "statement": "CREATE INDEX CONCURRENTLY ix_servednoticehistory_hashed_external_id ON servednoticehistory (hashed_external_id)",
            "type": "index",
        },
    ],
    "stagedresourceancestor": [
        # primary key index
        {
            "name": "ix_staged_resource_ancestor_pkey",
            "statement": "CREATE UNIQUE INDEX CONCURRENTLY ix_staged_resource_ancestor_pkey ON stagedresourceancestor (id)",
            "type": "index",
        },
        # unique constraint + index
        {
            "name": "ix_staged_resource_ancestor_unique",
            "statement": "CREATE UNIQUE INDEX CONCURRENTLY ix_staged_resource_ancestor_unique ON stagedresourceancestor (ancestor_urn, descendant_urn)",
            "type": "index",
            "constraint_name": "uq_staged_resource_ancestor",
        },
        {
            "name": "uq_staged_resource_ancestor",
            "statement": "ALTER TABLE stagedresourceancestor ADD CONSTRAINT uq_staged_resource_ancestor UNIQUE USING INDEX ix_staged_resource_ancestor_unique",
            "type": "constraint",
        },
        # foreign key constraints
        {
            "name": "fk_staged_resource_ancestor_ancestor",
            "statement": "ALTER TABLE stagedresourceancestor ADD CONSTRAINT fk_staged_resource_ancestor_ancestor FOREIGN KEY (ancestor_urn) REFERENCES stagedresource (urn) ON DELETE CASCADE",
            "type": "constraint",
        },
        {
            "name": "fk_staged_resource_ancestor_descendant",
            "statement": "ALTER TABLE stagedresourceancestor ADD CONSTRAINT fk_staged_resource_ancestor_descendant FOREIGN KEY (descendant_urn) REFERENCES stagedresource (urn) ON DELETE CASCADE",
            "type": "constraint",
        },
        # column indexes
        {
            "name": "ix_staged_resource_ancestor_ancestor",
            "statement": "CREATE INDEX CONCURRENTLY ix_staged_resource_ancestor_ancestor ON stagedresourceancestor (ancestor_urn)",
            "type": "index",
        },
        {
            "name": "ix_staged_resource_ancestor_descendant",
            "statement": "CREATE INDEX CONCURRENTLY ix_staged_resource_ancestor_descendant ON stagedresourceancestor (descendant_urn)",
            "type": "index",
        },
    ],
    "stagedresource": [
        {
            "name": "ix_stagedresource_monitor_config_resource_type_consent",
            "statement": "CREATE INDEX CONCURRENTLY ix_stagedresource_monitor_config_resource_type_consent ON stagedresource (monitor_config_id, resource_type, (meta->>'consent_aggregated'))",
            "type": "index",
        },
        {
            "name": "ix_stagedresource_system_vendor_consent",
            "statement": "CREATE INDEX CONCURRENTLY ix_stagedresource_system_vendor_consent ON stagedresource (system_id, vendor_id, (meta->>'consent_aggregated'))",
            "type": "index",
        },
    ],
}


def check_object_exists(db: Session, object_name: str) -> bool:
    """Checks Postgres' system catalogs to verify the existence of a given index or constraint in a specific table."""
    object_query = text(
        """
        SELECT EXISTS (
            SELECT 1
            FROM pg_indexes
            WHERE indexname = :object_name
        ) OR EXISTS (
            SELECT 1
            FROM pg_constraint
            WHERE conname = :object_name
        )
        """
    )
    result = db.execute(
        object_query,
        {
            "object_name": object_name,
        },
    ).scalar()
    return result


def create_object(db: Session, object_statement: str, object_name: str) -> None:
    """Executes the index or constraint creation statement."""
    logger.info(f"Creating index/constraint object: '{object_name}'...")
    with db.bind.connect().execution_options(
        isolation_level="AUTOCOMMIT"
    ) as connection:
        connection.execute(text(object_statement))
    logger.info(f"Successfully created index/constraint object: '{object_name}'")


def check_and_create_objects(
    db: Session, table_object_map: Dict[str, List[Dict[str, str]]], lock: Lock
) -> Dict[str, str]:
    """Returns a dictionary of any indices or constraints that were created."""
    object_info: Dict[str, str] = {}
    for _, objects in table_object_map.items():
        for object_data in objects:
            object_name = object_data["name"]
            object_statement = object_data["statement"]
            object_type = object_data["type"]
            object_exists = check_object_exists(db, object_name)

            if not object_exists:
                if object_type == "index":
                    constraint_name = object_data.get("constraint_name")
                    if constraint_name:
                        constraint_exists = check_object_exists(db, constraint_name)
                        if constraint_exists:
                            logger.debug(
                                f"Constraint {constraint_name} already exists, skipping index creation for {object_name}"
                            )
                            continue

                create_object(db, object_statement, object_name)
                object_info[object_name] = "created"
            else:
                logger.debug(
                    f"Object {object_name} already exists, skipping index/constraint creation"
                )
            lock.reacquire()

    return object_info


# Lock key for the post upgrade index creation
POST_UPGRADE_INDEX_CREATION_LOCK = "post_upgrade_index_creation_lock"
# The timeout of the lock for the post upgrade index creation, in seconds
POST_UPGRADE_INDEX_CREATION_LOCK_TIMEOUT_SECONDS = 600  # 10 minutes


def post_upgrade_index_creation_task() -> None:
    """
    Task for creating indices and constraints that were deferred
    as part of a standard Fides/Alembic migration.

    This task is to be kicked off as a background task during application startup,
    after standard migrations have been applied.

    If all specified indexes and constraints are created, the task is effectively
    a no-op, as it checks for the presence of the objects in the database before
    creating them.
    """
    with redis_lock(
        POST_UPGRADE_INDEX_CREATION_LOCK,
        POST_UPGRADE_INDEX_CREATION_LOCK_TIMEOUT_SECONDS,
    ) as lock:
        if not lock:
            return

        SessionLocal = get_db_session(CONFIG)
        with SessionLocal() as db:
            object_info: Dict[str, str] = check_and_create_objects(
                db, TABLE_OBJECT_MAP, lock
            )
            if object_info:
                logger.info(
                    f"Post upgrade index creation output: {json.dumps(object_info)}"
                )
            else:
                logger.debug("All indices and constraints already created")


def initiate_post_upgrade_index_creation() -> None:
    """Initiates scheduler to migrate all non-credential tables using a bcrypt hash to use a SHA-256 hash"""

    if CONFIG.test_mode:
        logger.debug("Skipping post upgrade index creation in test mode")
        return

    assert (
        scheduler.running
    ), "Scheduler is not running! Cannot migrate tables with bcrypt hashes."

    logger.info("Initiating scheduler for hash migration")
    scheduler.add_job(
        func=post_upgrade_index_creation_task,
        id=POST_UPGRADE_INDEX_CREATION,
    )



from typing import Any, Dict, List, Optional, Union

import pydash
from loguru import logger
from requests import Response

from fides.api.models.privacy_request import PrivacyRequest
from fides.api.schemas.saas.strategy_configuration import (
    ExtractForExecutionLogPostProcessorConfiguration,
)
from fides.api.service.execution_context import add_execution_log_message
from fides.api.service.processors.post_processor_strategy.post_processor_strategy import (
    PostProcessorStrategy,
)


class ExtractForExecutionLogPostProcessorStrategy(PostProcessorStrategy):
    """Postprocessor that extracts data from API response contents and adds it to execution log messages"""

    name = "extract_for_execution_log"
    configuration_model = ExtractForExecutionLogPostProcessorConfiguration

    def __init__(self, configuration: ExtractForExecutionLogPostProcessorConfiguration):
        self.configuration = configuration

    def process(
        self,
        data: Union[List[Dict[str, Any]], Dict[str, Any]],
        identity_data: Optional[Dict[str, Any]] = None,
        privacy_request: Optional[PrivacyRequest] = None,
        response: Optional[Response] = None,
    ) -> Union[List[Dict[str, Any]], Dict[str, Any]]:
        """
        Extract data from response contents and add to execution log messages.
        Data is passed through unchanged (non-destructive postprocessor).
        """
        config = self.configuration

        try:
            # Only process if response.contents is available
            if response is None or not response.content:
                return data

            if config.path is None:
                # No path specified - use entire response contents as string
                message = str(response.content)
            else:
                # Attempt to parse the response content as JSON before extracting
                try:
                    json_data: Any = response.json()
                except ValueError:
                    logger.debug(
                        "Unable to parse response content as JSON for path extraction"
                    )
                    return data

                extracted_value = pydash.get(json_data, config.path)
                if extracted_value is None:
                    logger.debug(f"No data found at path '{config.path}'")
                    return data

                message = str(extracted_value)

            # Add message to execution context
            add_execution_log_message(message)
            logger.info(message)

        except Exception as e:
            error_msg = f"Error in extract_for_execution_log postprocessor: {e}"
            logger.error(error_msg)
            add_execution_log_message(error_msg)

        return data



"""add property path

Revision ID: d7feff990012
Revises: 0debabbb9c6a
Create Date: 2024-05-15 00:20:47.115547

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "d7feff990012"
down_revision = "0debabbb9c6a"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "plus_property_path",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("property_id", sa.String(), nullable=False),
        sa.Column("path", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["property_id"],
            ["plus_property.id"],
        ),
        sa.PrimaryKeyConstraint("id", "property_id"),
    )
    op.create_index(
        op.f("ix_plus_property_path_id"), "plus_property_path", ["id"], unique=False
    )
    op.create_index(
        op.f("ix_plus_property_path_path"), "plus_property_path", ["path"], unique=True
    )
    op.create_index(
        op.f("ix_plus_property_path_property_id"),
        "plus_property_path",
        ["property_id"],
        unique=False,
    )
    op.drop_column("plus_property", "paths")


def downgrade():
    op.add_column(
        "plus_property",
        sa.Column(
            "paths",
            postgresql.ARRAY(sa.VARCHAR()),
            server_default=sa.text("'{}'::character varying[]"),
            autoincrement=False,
            nullable=False,
        ),
    )
    op.drop_index(
        op.f("ix_plus_property_path_property_id"), table_name="plus_property_path"
    )
    op.drop_index(op.f("ix_plus_property_path_path"), table_name="plus_property_path")
    op.drop_index(op.f("ix_plus_property_path_id"), table_name="plus_property_path")
    op.drop_table("plus_property_path")



"""update consent mechanism

Revision ID: 144d9b85c712
Revises: 8188ddbf8cae
Create Date: 2023-04-12 14:57:07.532660

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "144d9b85c712"
down_revision = "8188ddbf8cae"
branch_labels = None
depends_on = None


def upgrade():
    """
    Extra steps added here to avoid the error that enums have to be committed before they can be used.  This
    avoids having to commit in the middle of this migration and lets the entire thing be completed in a single transaction
    """
    op.execute("alter type consentmechanism rename to consentmechanismold")

    op.execute(
        "create type consentmechanism as enum ('opt_in', 'opt_out', 'notice_only', 'necessary');"
    )
    op.execute(
        "create type consentmechanismnew as enum ('opt_in', 'opt_out', 'notice_only');"
    )
    op.execute(
        (
            "alter table privacynoticehistory alter column consent_mechanism type consentmechanism using "
            "consent_mechanism::text::consentmechanism"
        )
    )
    op.execute(
        (
            "alter table privacynotice alter column consent_mechanism type consentmechanism using "
            "consent_mechanism::text::consentmechanism"
        )
    )
    op.execute(
        (
            "update privacynoticehistory set consent_mechanism = 'notice_only' where consent_mechanism = 'necessary';"
        )
    )
    op.execute(
        (
            "update privacynotice set consent_mechanism = 'notice_only' where consent_mechanism = 'necessary';"
        )
    )
    op.execute(
        (
            "alter table privacynoticehistory alter column consent_mechanism type consentmechanismnew using "
            "consent_mechanism::text::consentmechanismnew"
        )
    )
    op.execute(
        (
            "alter table privacynotice alter column consent_mechanism type consentmechanismnew using "
            "consent_mechanism::text::consentmechanismnew"
        )
    )
    op.execute(("drop type consentmechanismold;"))
    op.execute(("drop type consentmechanism;"))
    op.execute(("alter type consentmechanismnew rename to consentmechanism;"))


def downgrade():
    op.execute("alter type consentmechanism rename to consentmechanismold")

    op.execute(
        "create type consentmechanism as enum ('opt_in', 'opt_out', 'notice_only', 'necessary');"
    )
    op.execute(
        "create type consentmechanismnew as enum ('opt_in', 'opt_out', 'necessary');"
    )
    op.execute(
        (
            "alter table privacynoticehistory alter column consent_mechanism type consentmechanism using "
            "consent_mechanism::text::consentmechanism"
        )
    )
    op.execute(
        (
            "alter table privacynotice alter column consent_mechanism type consentmechanism using "
            "consent_mechanism::text::consentmechanism"
        )
    )
    op.execute(
        (
            "update privacynoticehistory set consent_mechanism = 'necessary' where consent_mechanism = 'notice_only';"
        )
    )
    op.execute(
        (
            "update privacynotice set consent_mechanism = 'necessary' where consent_mechanism = 'notice_only';"
        )
    )
    op.execute(
        (
            "alter table privacynoticehistory alter column consent_mechanism type consentmechanismnew using "
            "consent_mechanism::text::consentmechanismnew"
        )
    )
    op.execute(
        (
            "alter table privacynotice alter column consent_mechanism type consentmechanismnew using "
            "consent_mechanism::text::consentmechanismnew"
        )
    )
    op.execute(("drop type consentmechanismold;"))
    op.execute(("drop type consentmechanism;"))
    op.execute(("alter type consentmechanismnew rename to consentmechanism;"))



"""
Define routes for the ctl-related objects.

All routes should get imported here and added to the CTL_ROUTER
"""

from fastapi import APIRouter

from .endpoints.generate import GENERATE_ROUTER
from .endpoints.generic import (
    DATA_CATEGORY_ROUTER,
    DATA_SUBJECT_ROUTER,
    DATA_USE_ROUTER,
    DATASET_ROUTER,
    EVALUATION_ROUTER,
    ORGANIZATION_ROUTER,
    POLICY_ROUTER,
)
from .endpoints.system import (
    SYSTEM_CONNECTION_INSTANTIATE_ROUTER,
    SYSTEM_CONNECTIONS_ROUTER,
    SYSTEM_ROUTER,
)
from .endpoints.validate import VALIDATE_ROUTER

# ADMIN and HEALTH routers are explicitly _excluded_ here.
# those depend directly on database modules, and therefore
# cause circular dependency problems when initialized as
# part of this module. see https://github.com/ethyca/fides/issues/3652
routers = [
    DATA_CATEGORY_ROUTER,
    DATA_SUBJECT_ROUTER,
    DATA_USE_ROUTER,
    DATASET_ROUTER,
    EVALUATION_ROUTER,
    GENERATE_ROUTER,
    ORGANIZATION_ROUTER,
    POLICY_ROUTER,
    SYSTEM_CONNECTION_INSTANTIATE_ROUTER,
    SYSTEM_CONNECTIONS_ROUTER,
    SYSTEM_ROUTER,
    VALIDATE_ROUTER,
]

CTL_ROUTER = APIRouter()
for router in routers:
    CTL_ROUTER.include_router(router)



from __future__ import annotations

import zipfile
from datetime import datetime
from io import BytesIO
from typing import Any, Generator, Tuple

from loguru import logger
from stream_zip import _ZIP_32_TYPE

from fides.api.service.storage.streaming.schemas import AttachmentProcessingInfo
from fides.api.service.storage.streaming.smart_open_client import SmartOpenStorageClient


def stream_dsr_buffer_to_storage(
    storage_client: SmartOpenStorageClient,
    bucket_name: str,
    file_key: str,
    dsr_buffer: BytesIO,
) -> None:
    """Stream DSR buffer to storage using smart-open streaming.

    This function handles only the storage streaming concern, accepting a pre-generated
    DSR buffer to maintain clear separation of concerns.

    Args:
        storage_client: The storage client for streaming uploads
        bucket_name: Storage bucket name
        file_key: File key in storage
        dsr_buffer: Pre-generated DSR report buffer (BytesIO)
    """
    # Get the content from the buffer
    content = dsr_buffer.getvalue()
    try:
        # Use smart-open's streaming upload for efficient memory usage
        with storage_client.stream_upload(bucket_name, file_key) as upload_stream:
            upload_stream.write(content)

    except Exception as e:
        logger.error("Failed to upload DSR report using smart-open streaming: {}", e)
        raise e


def create_dsr_report_files_generator(
    dsr_buffer: BytesIO,
    all_attachments: list["AttachmentProcessingInfo"],
    bucket_name: str,
    max_workers: int,
    batch_size: int,
) -> Generator[
    Tuple[str, datetime, int, Any, Generator[bytes, None, None]], None, None
]:
    """Create a ZIP generator for DSR report HTML files only.

    This method extracts and yields the HTML files from a DSR report buffer.
    Note: This function only handles the DSR report files (HTML, CSS, etc.).
    The caller is responsible for combining this with attachment files to create
    the complete ZIP.

    Args:
        dsr_buffer: The DSR report buffer (ZIP file as BytesIO)
        all_attachments: List of validated attachments (used for logging only)
        bucket_name: Storage bucket name (used for logging only)
        max_workers: Maximum parallel workers (used for logging only)
        batch_size: Number of attachments to process in each batch (used for logging only)

    Returns:
        Generator yielding DSR report files in stream_zip format
    """
    logger.debug(
        f"Creating DSR report files generator with {len(all_attachments)} attachments"
    )

    # Reset buffer position to ensure we can read from it
    dsr_buffer.seek(0)

    # Extract and yield the DSR report files from the buffer
    # The dsr_buffer is already a ZIP file, so we need to extract and re-yield its contents
    with zipfile.ZipFile(dsr_buffer) as dsr_zip:
        for file_info in dsr_zip.filelist:
            if not file_info.is_dir():
                # Read the file content and yield it in stream_zip format
                content = dsr_zip.read(file_info.filename)

                def content_generator(
                    file_content: bytes,
                ) -> Generator[bytes, None, None]:
                    yield file_content

                yield file_info.filename, datetime.now(), 0o644, _ZIP_32_TYPE(), content_generator(
                    content
                )

    logger.debug("DSR report files extracted and ready for ZIP creation")



"""add_location_groups_column

Revision ID: 0c65325843bd
Revises: a1e23b70f2b2
Create Date: 2024-02-21 19:56:39.228072

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "0c65325843bd"
down_revision = "a1e23b70f2b2"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "location_regulation_selections",
        sa.Column(
            "selected_location_groups",
            sa.ARRAY(sa.String()),
            server_default="{}",
            nullable=False,
        ),
    )


def downgrade():
    op.drop_column("location_regulation_selections", "selected_location_groups")



from typing import Any, Dict, List, Optional, Set, Union

from fideslang.models import FidesCollectionKey, FidesDatasetReference
from fideslang.validation import FidesKey
from pydantic import BaseModel, ConfigDict, field_validator, model_validator

from fides.api.common_exceptions import ValidationError
from fides.api.graph.config import (
    Collection,
    CollectionAddress,
    Field,
    FieldAddress,
    GraphDataset,
    ScalarField,
)
from fides.api.schemas.base_class import FidesSchema
from fides.api.schemas.limiter.rate_limit_config import RateLimitConfig
from fides.api.schemas.policy import ActionType
from fides.api.schemas.saas.display_info import SaaSDisplayInfo
from fides.api.schemas.saas.shared_schemas import HTTPMethod
from fides.api.service.saas_request.saas_request_override_factory import (
    SaaSRequestOverrideFactory,
    SaaSRequestType,
)


class ParamValue(BaseModel):
    """
    A named variable which can be sourced from identities, dataset references, or connector params. These values
    are used to replace the placeholders in the path, header, query, and body param values.
    """

    name: str
    identity: Optional[str] = None
    references: Optional[List[Union[FidesDatasetReference, str]]] = None
    connector_param: Optional[str] = None
    unpack: Optional[bool] = False

    @field_validator("references")
    @classmethod
    def check_reference_direction(
        cls, references: Optional[List[Union[FidesDatasetReference, str]]]
    ) -> Optional[List[Union[FidesDatasetReference, str]]]:
        """Validates the request_param only contains inbound references"""
        for reference in references or {}:
            if isinstance(reference, FidesDatasetReference):
                if reference.direction == "to":
                    raise ValueError(
                        "References can only have a direction of 'from', found 'to'"
                    )
        return references

    @model_validator(mode="after")
    def check_exactly_one_value_field(self) -> "ParamValue":
        value_fields = [
            bool(self.identity),
            bool(self.references),
            bool(self.connector_param),
        ]
        if sum(value_fields) != 1:
            raise ValueError(
                "Must have exactly one of 'identity', 'references', or 'connector_param'"
            )
        return self


class Strategy(BaseModel):
    """General shape for swappable strategies (ex: auth, processors, pagination, etc.)"""

    strategy: str
    configuration: Optional[Dict[str, Any]] = None


class ClientConfig(BaseModel):
    """Definition for an authenticated base HTTP client"""

    protocol: str
    host: str
    authentication: Optional[Strategy] = None


class Header(BaseModel):
    name: str
    value: str


class QueryParam(BaseModel):
    name: str
    value: Union[int, str]


class SaaSRequest(BaseModel):
    """
    A single request with static or dynamic path, headers, query, and body params.
    Also specifies the names and sources for the param values needed to build the request.

    Includes optional strategies for postprocessing and pagination.
    """

    request_override: Optional[str] = None
    path: Optional[str] = None
    method: Optional[HTTPMethod] = None
    headers: Optional[List[Header]] = []
    query_params: Optional[List[QueryParam]] = []
    body: Optional[str] = None
    param_values: Optional[List[ParamValue]] = []
    client_config: Optional[ClientConfig] = None
    data_path: Optional[str] = None
    postprocessors: Optional[List[Strategy]] = None
    pagination: Optional[Strategy] = None
    grouped_inputs: Optional[List[str]] = []
    ignore_errors: Optional[Union[bool, List[int]]] = False
    rate_limit_config: Optional[RateLimitConfig] = None
    async_config: Optional[Strategy] = None
    skip_missing_param_values: Optional[bool] = (
        False  # Skip instead of raising an exception if placeholders can't be populated in body
    )
    model_config = ConfigDict(
        from_attributes=True, use_enum_values=True, extra="forbid"
    )

    @model_validator(mode="after")
    def validate_request_for_pagination(self) -> "SaaSRequest":
        """
        Calls the appropriate validation logic for the request based on
        the specified pagination strategy. Passes in the raw value dict
        before any field validation.
        """

        # delay import to avoid cyclic-dependency error - We still ignore the pylint error
        from fides.api.service.pagination.pagination_strategy import (  # pylint: disable=R0401
            PaginationStrategy,
        )

        pagination = self.pagination
        if pagination is not None:
            pagination_strategy = PaginationStrategy.get_strategy(
                pagination.strategy, pagination.configuration
            )
            pagination_strategy.validate_request(self.model_dump(mode="json"))
        return self

    @model_validator(mode="after")
    def validate_grouped_inputs(self) -> "SaaSRequest":
        """Validate that grouped_inputs must reference fields from the same collection"""
        grouped_inputs = set(self.grouped_inputs or [])

        if grouped_inputs:
            param_values = self.param_values or []
            names = {param.name for param in param_values}

            if not grouped_inputs.issubset(names):
                raise ValueError(
                    "Grouped input fields must also be declared as param_values."
                )

            referenced_collections: List[str] = []
            for param in param_values:
                if param.name in grouped_inputs:
                    if not param.references and not param.identity:
                        raise ValueError(
                            "Grouped input fields must either be reference fields or identity fields."
                        )
                    if param.references:
                        # reference may be a str, in which case it's an external reference.
                        # since external references are parameterized via secrets,
                        # they cannot be resolved and checked at this point in the validation.
                        # so here we only perform the check if the reference is a FidesDatasetReference
                        if isinstance(param.references[0], FidesDatasetReference):
                            collect = param.references[0].field.split(".")[0]
                            referenced_collections.append(collect)
                        else:
                            raise ValueError(
                                "Grouped inputs do not currently support external dataset references"
                            )

            if len(set(referenced_collections)) != 1:
                raise ValueError(
                    "Grouped input fields must all reference the same collection."
                )

        return self

    @model_validator(mode="after")
    def validate_request(self) -> "SaaSRequest":
        """Validate that configs related to request overrides are set properly"""
        if not self.request_override:
            if not self.path:
                raise ValueError(
                    "A request must specify a path if no request_override is provided"
                )
            if not self.method:
                raise ValueError(
                    "A request must specify a method if no request_override is provided"
                )

        else:  # if a request override is specified, many fields are NOT allowed
            invalid = [
                k
                for k in self.model_fields
                if getattr(self, k)
                and k not in ("request_override", "param_values", "grouped_inputs")
            ]
            if invalid:
                invalid_joined = ", ".join(invalid)
                raise ValueError(
                    f"Invalid properties [{invalid_joined}] on a request with request_override specified"
                )

        return self


class ReadSaaSRequest(SaaSRequest):
    """
    An extension of the base SaaSRequest that allows the inclusion of an output template
    that is used to format each collection result.
    """

    output: Optional[str] = None

    @model_validator(mode="after")
    def validate_request(self) -> "ReadSaaSRequest":
        """Validate that configs related to read requests are set properly"""
        if not self.request_override:
            if not (self.path or self.output):
                raise ValueError(
                    "A read request must specify either a path or an output if no request_override is provided"
                )
            if self.path and not self.method:
                raise ValueError(
                    "A read request must specify a method if a path is provided and no request_override is specified"
                )
        else:
            allowed_fields = {
                "request_override",
                "param_values",
                "grouped_inputs",
            }
            invalid = [
                k
                for k in self.model_fields
                if getattr(self, k) and k not in allowed_fields
            ]
            if invalid:
                invalid_joined = ", ".join(invalid)
                raise ValueError(
                    f"Invalid properties [{invalid_joined}] on a read request with request_override specified"
                )
        return self


class SaaSRequestMap(BaseModel):
    """A map of actions to SaaS requests"""

    read: Union[ReadSaaSRequest, List[ReadSaaSRequest]] = []
    update: Optional[SaaSRequest] = None
    delete: Optional[SaaSRequest] = None


class ConsentRequestMap(BaseModel):
    """A map of actions to Consent requests"""

    opt_in: Union[SaaSRequest, List[SaaSRequest]] = []
    opt_out: Union[SaaSRequest, List[SaaSRequest]] = []

    @field_validator("opt_in", "opt_out")
    @classmethod
    def validate_list_field(
        cls,
        field_value: Union[SaaSRequest, List[SaaSRequest]],
    ) -> List[SaaSRequest]:
        """Convert all opt_in/opt_out request formats to a list of requests.

        We allow either a single request or a list of requests to be defined, but this makes
        sure that everything is a list once that data has been read in.
        """
        if isinstance(field_value, SaaSRequest):
            return [field_value]
        return field_value


class Endpoint(BaseModel):
    """A collection of read/update/delete requests which corresponds to a FidesDataset collection (by name)"""

    name: str
    requests: SaaSRequestMap
    skip_processing: bool = False
    after: List[FidesCollectionKey] = []
    erase_after: List[FidesCollectionKey] = []

    @field_validator("requests")
    @classmethod
    def validate_grouped_inputs(
        cls,
        requests: SaaSRequestMap,
    ) -> SaaSRequestMap:
        """Validate that grouped_inputs are the same for every read request"""

        read_requests = requests.read
        if isinstance(read_requests, list) and len(read_requests) > 1:
            first, *rest = read_requests
            if not all(
                request.grouped_inputs == first.grouped_inputs for request in rest
            ):
                raise ValueError(
                    "The grouped_input values for every read request must be the same"
                )
        return requests


class ConnectorParam(BaseModel):
    """Used to define the required parameters for the connector (user and constants)"""

    name: str
    label: Optional[str] = None
    options: Optional[List[str]] = (
        None  # list of possible values for the connector param
    )
    # If you change the default_value type, update this in SaaSSchemaFactory.get_saas_schema
    default_value: Optional[Union[str, List[str], int, List[int]]] = None
    multiselect: Optional[bool] = False
    description: Optional[str] = None
    sensitive: Optional[bool] = False

    @model_validator(mode="before")
    @classmethod
    def validate_connector_param(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Verify the default_value is one of the values specified in the options list"""

        name = values.get("name")
        options: Optional[List[str]] = values.get("options")
        default_value: Optional[Union[str, List[str]]] = values.get("default_value")
        multiselect: Optional[bool] = values.get("multiselect")

        if options:
            if isinstance(default_value, str) and default_value not in options:
                raise ValueError(
                    f"'{default_value}' is not a valid option, default_value must be a value from [{', '.join(options)}]"
                )
            if isinstance(default_value, list):
                if not multiselect:
                    raise ValueError(
                        f"The default_value for the {name} connector_param must be a single value when multiselect is not enabled, not a list"
                    )

                invalid_options = [
                    value for value in default_value if value not in options
                ]
                if invalid_options:
                    raise ValueError(
                        f"[{', '.join(invalid_options)}] are not valid options, default_value must be a list of values from [{', '.join(options)}]"
                    )

        if multiselect and not options:
            raise ValueError(
                f"The 'multiselect' field in the {name} connector_param must be accompanied by an 'options' field containing a list of values."
            )

        return values


class ExternalDatasetReference(BaseModel):
    name: str
    label: Optional[str] = None
    description: Optional[str] = None


class SaaSConfigBase(BaseModel):
    """
    Used to store base info for a SaaS config
    """

    fides_key: FidesKey
    name: str
    type: str

    @property
    def fides_key_prop(self) -> FidesKey:
        return self.fides_key

    @property
    def name_prop(self) -> str:
        return self.name

    @field_validator("type", mode="before")
    @classmethod
    def lowercase_saas_type(cls, value: str) -> str:
        """Enforce lowercase on saas type."""
        return value.lower()

    model_config = ConfigDict(use_enum_values=True)


class SaaSConfig(SaaSConfigBase):
    """
    Used to store endpoint and param configurations for a SaaS connector.
    This is done to separate the details of how to make the API calls
    from the data provided by a given API collection.

    The required fields for the config are converted into a Dataset which is
    merged with the standard Fidesops Dataset to provide a complete set of dependencies
    for the graph traversal.
    """

    description: str
    version: str
    replaceable: bool = False
    connector_params: List[ConnectorParam]
    external_references: Optional[List[ExternalDatasetReference]] = None
    client_config: ClientConfig
    endpoints: List[Endpoint]
    test_request: SaaSRequest
    data_protection_request: Optional[SaaSRequest] = None  # GDPR Delete
    rate_limit_config: Optional[RateLimitConfig] = None
    consent_requests: Optional[ConsentRequestMap] = None
    user_guide: Optional[str] = None
    display_info: Optional[SaaSDisplayInfo] = None

    @property
    def top_level_endpoint_dict(self) -> Dict[str, Endpoint]:
        """Returns a map of endpoint names mapped to Endpoints"""
        return {endpoint.name: endpoint for endpoint in self.endpoints}

    def get_graph(self, secrets: Dict[str, Any]) -> GraphDataset:
        """Converts endpoints to a Dataset with collections and field references"""
        collections = []
        for endpoint in self.endpoints:
            fields: List[Field] = []

            read_requests: List[ReadSaaSRequest] = []
            if endpoint.requests.read:
                read_requests = (
                    endpoint.requests.read
                    if isinstance(endpoint.requests.read, list)
                    else [endpoint.requests.read]
                )

            delete_request = endpoint.requests.delete

            for read_request in read_requests:
                self._process_param_values(fields, read_request.param_values, secrets)

            if not read_requests and delete_request:
                # If the endpoint only specifies a delete request without a read,
                # then we must use the delete request's param_values instead.
                # One of the fields must automatically be flagged as a primary key
                # in order for the deletion to execute. See fides#1199
                self._process_param_values(fields, delete_request.param_values, secrets)
                if fields:
                    fields[0].primary_key = True

            if fields:
                grouped_inputs: Set[str] = set()
                if read_requests:
                    # the endpoint validator enforces that grouped inputs are the same
                    # for all read requests so we can just take the first one
                    grouped_inputs = set(read_requests[0].grouped_inputs or [])
                collections.append(
                    Collection(
                        name=endpoint.name,
                        fields=fields,
                        grouped_inputs=grouped_inputs,
                        skip_processing=endpoint.skip_processing,  # Sets this value for use when datasets are merged. If True, this will be later be removed.
                        after={
                            CollectionAddress(*s.split(".")) for s in endpoint.after
                        },
                        erase_after={
                            CollectionAddress(*s.split("."))
                            for s in endpoint.erase_after
                        },
                    )
                )

        return GraphDataset(
            name=super().name_prop,
            collections=collections,
            connection_key=super().fides_key_prop,
        )

    def _process_param_values(
        self,
        fields: List[Field],
        param_values: Optional[List[ParamValue]],
        secrets: Dict[str, Any],
    ) -> None:
        """
        Converts param values to dataset fields with identity and dataset references
        """
        for param in param_values or []:
            if param.references:
                references = []
                for reference in param.references:
                    resolved_reference = self.resolve_param_reference(
                        reference, secrets
                    )
                    first, *rest = resolved_reference.field.split(".")
                    references.append(
                        (
                            FieldAddress(resolved_reference.dataset, first, *rest),
                            (
                                # Convert back into a literal for Pydantic V2
                                resolved_reference.direction.value
                                if resolved_reference.direction
                                else resolved_reference.direction
                            ),
                        )
                    )
                fields.append(ScalarField(name=param.name, references=references))
            if param.identity:
                fields.append(ScalarField(name=param.name, identity=param.identity))

    @staticmethod
    def resolve_param_reference(
        reference: Union[str, FidesDatasetReference], secrets: Dict[str, Any]
    ) -> FidesDatasetReference:
        """
        If needed, resolves the given `reference` using the provided `secrets` `dict`.
        For ease of use, the given `reference` can either be a `str` or `FidesDatasetReference`,
        since a `ParamValue`'s `reference` may be of either type.

        If the `reference` is a `str`, then it's used as a key look up a value in the provided secrets dict,
        and a `FidesDatasetReference` is created and returned from the retrieved secrets object.

        If the `reference` is a `FidesDatasetReference`, then it's just returned as-is.
        """
        if isinstance(reference, str):
            if reference not in secrets.keys():
                raise ValidationError(
                    f"External dataset reference with provided name {reference} not found in connector's secrets."
                )
            reference = FidesDatasetReference.model_validate(secrets[reference])
        return reference

    @property
    def supported_actions(self) -> List[ActionType]:
        """Returns a list containing the privacy actions supported by the SaaS config."""

        supported_actions = []

        # check for access
        if any(
            requests.read
            for requests in [endpoint.requests for endpoint in self.endpoints]
        ):
            supported_actions.append(ActionType.access)

        # check for erasure
        if (
            any(
                request.update or request.delete
                for request in [endpoint.requests for endpoint in self.endpoints]
            )
            or self.data_protection_request
        ):
            supported_actions.append(ActionType.erasure)

        # consent is supported if the SaaSConfig has consent_requests defined
        # or if the SaaSConfig.type has an UPDATE_CONSENT function
        # registered in the SaaSRequestOverrideFactory
        if (
            self.consent_requests
            or self.type
            in SaaSRequestOverrideFactory.registry[
                SaaSRequestType.UPDATE_CONSENT
            ].keys()
        ):
            supported_actions.append(ActionType.consent)

        return supported_actions


class SaaSConfigValidationDetails(FidesSchema):
    """
    Message with any validation issues with the SaaS config
    """

    msg: Optional[str] = None


class ValidateSaaSConfigResponse(FidesSchema):
    """
    Response model for validating a SaaS config, which includes both the SaaS config
    itself (if valid) plus a details object describing any validation errors.
    """

    saas_config: SaaSConfig
    validation_details: SaaSConfigValidationDetails



# pylint: disable=unused-import
from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Any, List

from citext import CIText
from sqlalchemy import Boolean, Column, DateTime
from sqlalchemy import Enum as EnumColumn
from sqlalchemy import String
from sqlalchemy.orm import Session, relationship
from sqlalchemy_utils.types.encrypted.encrypted_type import (
    AesGcmEngine,
    StringEncryptedType,
)

from fides.api.common_exceptions import SystemManagerException
from fides.api.cryptography.cryptographic_util import (
    generate_salt,
    hash_credential_with_salt,
)
from fides.api.db.base_class import Base
from fides.api.models.audit_log import AuditLog

# Intentionally importing SystemManager here to build the FidesUser.systems relationship
from fides.api.schemas.user import DisabledReason
from fides.config import CONFIG

if TYPE_CHECKING:
    from fides.api.models.fides_user_permissions import FidesUserPermissions
    from fides.api.models.fides_user_respondent_email_verification import (
        FidesUserRespondentEmailVerification,
    )
    from fides.api.models.sql_models import System  # type: ignore[attr-defined]
    from fides.api.models.system_manager import SystemManager


class FidesUser(Base):
    """The DB ORM model for FidesUser."""

    username = Column(CIText, unique=True, index=True)
    email_address = Column(CIText, unique=True, nullable=True)
    first_name = Column(String, nullable=True)
    last_name = Column(String, nullable=True)
    hashed_password = Column(String, nullable=True)
    salt = Column(String, nullable=True)
    disabled = Column(Boolean, nullable=False, server_default="f")
    disabled_reason = Column(EnumColumn(DisabledReason), nullable=True)
    last_login_at = Column(DateTime(timezone=True), nullable=True)
    password_reset_at = Column(DateTime(timezone=True), nullable=True)
    password_login_enabled = Column(Boolean, nullable=True)
    totp_secret = Column(
        StringEncryptedType(
            type_in=String(),
            key=CONFIG.security.app_encryption_key,
            engine=AesGcmEngine,
            padding="pkcs5",
        ),
        nullable=True,
    )

    # passive_deletes="all" prevents audit logs from having their
    # privacy_request_id set to null when a privacy_request is deleted.
    # We want to retain for record-keeping.
    audit_logs = relationship(
        AuditLog,
        backref="fides_user",
        lazy="dynamic",
        passive_deletes="all",
        primaryjoin="foreign(AuditLog.user_id)==FidesUser.id",
    )

    client = relationship(  # type: ignore
        "ClientDetail", backref="user", cascade="all, delete", uselist=False
    )

    systems = relationship("System", secondary="systemmanager", back_populates="data_stewards")  # type: ignore
    # permissions relationship is defined via backref in FidesUserPermissions
    email_verifications = relationship(
        "FidesUserRespondentEmailVerification",
        back_populates="user",
        cascade="all,delete",
        lazy="dynamic",
        foreign_keys="[FidesUserRespondentEmailVerification.user_id]",
    )
    permissions = relationship(
        "FidesUserPermissions",
        back_populates="user",
        cascade="all,delete",
        uselist=False,
    )

    @property
    def system_ids(self) -> List[str]:
        return [system.id for system in self.systems]

    @classmethod
    def hash_password(cls, password: str, encoding: str = "UTF-8") -> tuple[str, str]:
        """Utility function to hash a user's password with a generated salt"""
        salt = generate_salt()
        hashed_password = hash_credential_with_salt(
            password.encode(encoding),
            salt.encode(encoding),
        )
        return hashed_password, salt

    @classmethod
    def create(
        cls, db: Session, data: dict[str, Any], check_name: bool = True
    ) -> FidesUser:
        """Create a FidesUser by hashing the password with a generated salt
        and storing the hashed password and the salt"""
        hashed_password = None
        salt = None

        if password := data.get("password"):
            hashed_password, salt = FidesUser.hash_password(password)

        user = super().create(
            db,
            data={
                "salt": salt,
                "hashed_password": hashed_password,
                "username": data["username"],
                "email_address": data.get("email_address"),
                "first_name": data.get("first_name"),
                "last_name": data.get("last_name"),
                "disabled": data.get("disabled") or False,
                "disabled_reason": data.get("disabled_reason"),
                "password_login_enabled": data.get("password_login_enabled"),
            },
            check_name=check_name,
        )

        return user  # type: ignore

    @classmethod
    def create_respondent(cls, db: Session, data: dict[str, Any]) -> FidesUser:
        """Create an external respondent user. This user will not be able to login with a password and
        requires an email address to be provided.
        """
        if not data.get("email_address"):
            raise ValueError("Email address is required for external respondents")
        if data.get("password"):
            raise ValueError("Password login is not allowed for external respondents")
        data["password_login_enabled"] = False
        return cls.create(db, data)

    def credentials_valid(self, password: str, encoding: str = "UTF-8") -> bool:
        """Verifies that the provided password is correct."""
        if self.salt is None:
            return False

        provided_password_hash = hash_credential_with_salt(
            password.encode(encoding),
            self.salt.encode(encoding),
        )

        return provided_password_hash == self.hashed_password

    def update_password(self, db: Session, new_password: str) -> None:
        """Updates the user's password to the specified value.

        No validations are performed on the old/existing password within this function.
        """
        if self.permissions is not None:
            if self.permissions.is_external_respondent():
                raise ValueError(
                    "Password changes are not allowed for external respondents"
                )

        hashed_password, salt = FidesUser.hash_password(new_password)
        self.hashed_password = hashed_password  # type: ignore
        self.salt = salt  # type: ignore
        self.password_reset_at = datetime.utcnow()  # type: ignore
        self.save(db)

    def update_email_address(self, db: Session, new_email_address: str) -> None:
        """Updates the user's email address to the specified value."""
        if self.permissions is not None:
            if self.permissions.is_external_respondent():
                raise ValueError(
                    "Email address changes are not allowed for external respondents"
                )

        self.email_address = new_email_address  # type: ignore
        self.save(db)

    def set_as_system_manager(self, db: Session, system: System) -> None:
        """Add a user as one of the system managers for the given system
        If applicable, also update the systems on the user's client
        """
        if (
            not type(system).__name__ == "System"
        ):  # Checking type instead of with isinstance to avoid import errors
            raise SystemManagerException(
                "Must pass in a system to set user as system manager."
            )

        if self in system.data_stewards:
            raise SystemManagerException(
                f"User '{self.username}' is already a system manager of '{system.name}'."
            )

        if self.permissions is not None:
            if self.permissions.is_external_respondent():
                raise SystemManagerException(
                    "External respondents cannot be system managers."
                )

        self.systems.append(system)
        self.save(db=db)

        if self.client:
            self.client.update(db=db, data={"systems": self.system_ids})

    def remove_as_system_manager(self, db: Session, system: System) -> None:
        """Remove a user as one of the system managers for the given system
        If applicable, also update the systems on the user's client
        """
        if not type(system).__name__ == "System":
            raise SystemManagerException(
                "Must pass in a system to remove user as system manager."
            )

        try:
            self.systems.remove(system)
            self.save(db=db)
        except ValueError:
            raise SystemManagerException(
                f"User '{self.username}' is not a manager of system '{system.name}'."
            )

        if self.client:
            self.client.update(db=db, data={"systems": self.system_ids})



"""Add WebMonitorGroupJob model

Revision ID: 30369bb8ae01
Revises: 67d0e389b003
Create Date: 2025-09-04 13:18:41.306744

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "30369bb8ae01"
down_revision = "67d0e389b003"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "web_monitor_group_job",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("group_id", sa.String(), nullable=False, unique=True),
        sa.Column(
            "raw_experiences_data",
            postgresql.JSONB(astext_type=sa.Text()),
            nullable=True,
        ),
        sa.Column(
            "processed_experiences_data",
            postgresql.JSONB(astext_type=sa.Text()),
            nullable=True,
        ),
        sa.Column("is_test_run", sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("group_id"),
    )
    op.create_index(
        op.f("ix_web_monitor_group_job_id"),
        "web_monitor_group_job",
        ["id"],
        unique=False,
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(
        op.f("ix_web_monitor_group_job_id"), table_name="web_monitor_group_job"
    )
    op.drop_table("web_monitor_group_job")
    # ### end Alembic commands ###



import json
from typing import Any, Dict, List, NamedTuple, Optional, Tuple

from fideslang.models import FidesDatasetReference
from loguru import logger
from sqlalchemy.exc import InternalError, OperationalError
from sqlalchemy.orm import Query, Session

from fides.api.common_exceptions import MessageDispatchException
from fides.api.graph.config import CollectionAddress, GraphDataset
from fides.api.graph.execution import ExecutionNode
from fides.api.models.connectionconfig import ConnectionConfig, ConnectionTestStatus
from fides.api.models.datasetconfig import DatasetConfig
from fides.api.models.privacy_request import (
    ExecutionLog,
    PrivacyRequest,
    RequestTask,
    TraversalDetails,
)
from fides.api.models.worker_task import ExecutionLogStatus
from fides.api.schemas.connection_configuration.connection_secrets_dynamic_erasure_email import (
    DynamicErasureEmailSchema,
)
from fides.api.schemas.policy import ActionType
from fides.api.service.connectors.base_connector import BaseConnector
from fides.api.service.connectors.base_erasure_email_connector import (
    BaseErasureEmailConnector,
    filter_user_identities_for_connector,
    send_single_erasure_email,
)
from fides.config import get_config

CONFIG = get_config()


class DynamicErasureEmailConnectorException(Exception):
    pass


class ProcessedConfig(NamedTuple):
    """
    Helper type to make the code easier to read, returned by process_connector_config
    and used by get_email_and_vendor_from_custom_request_fields.
    """

    graph_dataset: GraphDataset
    connector: BaseConnector
    collection_address: CollectionAddress
    collection_data: Any
    email_field: str
    vendor_field: str
    dsr_field_to_collection_field: Dict[str, str]


class BatchedIdentitiesData(NamedTuple):
    email_address: str
    vendor_name: str
    identities: List[str]


class DynamicErasureEmailConnector(BaseErasureEmailConnector):
    """Email Erasure Connector that performs a lookup for the email to use."""

    config: DynamicErasureEmailSchema

    def get_config(self, configuration: ConnectionConfig) -> DynamicErasureEmailSchema:
        return DynamicErasureEmailSchema(**configuration.secrets or {})

    def batch_email_send(self, privacy_requests: Query, batch_id: str) -> None:
        logger.debug(
            "Starting batch erasure email {} for connector: {} ...",
            batch_id,
            self.configuration.key,
        )

        db: Session = Session.object_session(self.configuration)
        processed_config = self.process_connector_config(db, privacy_requests)

        skipped_privacy_requests: List[str] = []

        # We'll batch identities for each (email address, vendor_name) pair
        # the keys of the dictionary will be email_address:vendor_name
        batched_identities: Dict[str, BatchedIdentitiesData] = {}

        for privacy_request in privacy_requests:
            try:
                custom_field_data = (
                    self.get_email_and_vendor_from_custom_request_fields(
                        db,
                        processed_config,
                        privacy_request,
                    )
                )

                # if we couldn't retrieve the email address / vendor, we skip the privacy request
                # (error was already logged within the function, so we just continue to the
                # next privacy request)
                if not custom_field_data:
                    continue

                user_identities: Dict[str, Any] = (
                    privacy_request.get_cached_identity_data()
                )
                filtered_user_identities: Dict[str, Any] = (
                    filter_user_identities_for_connector(self.config, user_identities)
                )

                # If there are no user identities, we just skip the privacy request
                if not filtered_user_identities:
                    skipped_privacy_requests.append(privacy_request.id)
                    self.add_skipped_log(db, privacy_request)
                    continue

                email, vendor_name = custom_field_data

                key = f"{email}:{vendor_name}"
                if key not in batched_identities:
                    batched_identities[key] = BatchedIdentitiesData(
                        email, vendor_name, []
                    )

                batched_identities[key].identities.extend(
                    filtered_user_identities.values()
                )

            except Exception as exc:
                logger.error(
                    "An error occurred when retrieving email from custom request fields for connector {}. Erasure email for privacy request with id {} not sent. Error: {}",
                    self.configuration.key,
                    privacy_request.id,
                    exc,
                )
                self.error_privacy_request(db, privacy_request, str(exc))

        if not batched_identities:
            logger.info(
                "Skipping erasure email send for connector '{}' and batch '{}'. "
                "No corresponding user identities or email addresses found for pending privacy requests.",
                self.configuration.key,
                batch_id,
            )
            return

        logger.info(
            "Sending batch erasure email {} for connector {}...",
            batch_id,
            self.configuration.key,
        )

        # Don't care about the keys in the dict, just the values
        for batched_identity_data in batched_identities.values():
            try:
                send_single_erasure_email(
                    db=db,
                    subject_email=batched_identity_data.email_address,
                    subject_name=batched_identity_data.vendor_name,
                    batch_identities=batched_identity_data.identities,
                    test_mode=False,
                )
            except MessageDispatchException as exc:
                message = f"Dynamic batch erasure email {batch_id} for connector {self.configuration.key} failed with exception {exc}"
                logger.error(message)
                self.error_all_privacy_requests(
                    db,
                    privacy_requests.filter(
                        PrivacyRequest.id.notin_(skipped_privacy_requests)
                    ),
                    message,
                )
                raise exc

        # create an audit event for each privacy request ID
        for privacy_request in privacy_requests:
            if privacy_request.id not in skipped_privacy_requests:
                ExecutionLog.create(
                    db=db,
                    data={
                        "connection_key": self.configuration.key,
                        "dataset_name": self.configuration.name_or_key,
                        "collection_name": self.configuration.name_or_key,
                        "privacy_request_id": privacy_request.id,
                        "action_type": ActionType.erasure,
                        "status": ExecutionLogStatus.complete,
                        "message": f"Erasure email instructions dispatched for '{self.configuration.name_or_key}'",
                    },
                )

    def get_collection_and_field_from_reference(
        self,
        db: Session,
        privacy_requests: Query,
        dataset_reference: FidesDatasetReference,
    ) -> Tuple[str, str]:
        dataset_key = dataset_reference.dataset  # pylint: disable=no-member

        # Field should be of the form collection_name.field_name, potentially with nested fields
        # e.g collection_name.field_name.nested_field_name, so splitting by "." should give
        # a list of length greater than or equal to 2.
        split_field = dataset_reference.field.split(".")  # pylint: disable=no-member

        if len(split_field) < 2:
            field = dataset_reference.field  # pylint: disable=no-member
            error_log = f"Invalid dataset reference field {field} for dataset {dataset_key}. Failed to send dynamic erasure emails for connector: {self.configuration.key}."
            logger.error(error_log)
            self.error_all_privacy_requests(
                db,
                privacy_requests,
                f"Connector configuration references invalid dataset field {field} for dataset {dataset_key}.",
            )
            raise DynamicErasureEmailConnectorException(error_log)

        # Get collection name and field name
        collection_name = split_field[0]
        field_name = ".".join(split_field[1:])

        return collection_name, field_name

    def process_connector_config(
        self, db: Session, privacy_requests: Query
    ) -> ProcessedConfig:
        """
        Validates and processes the connector config. Returns a ProcessedConfig that has the graph_dataset
        corresponding to the referenced dataset config, the associated connector instance (different from the
        DynamicErasureEmailConnector), and the collection address and fields of the dataset where we need to
        lookup the recipient email address and vendor name.

        Raises a DynamicErasureEmailConnectorException if the connector configuration is invalid.
        """
        # Import here to avoid circular import error
        # TODO: once we incorporate custom request fields into the DSR graph logic  we won't need this import here,
        # so ignoring cyclic import warning for now
        from fides.api.service.connectors import (  # pylint: disable=cyclic-import
            get_connector,
        )

        email_dataset_reference = self.config.recipient_email_address
        email_dataset_key = email_dataset_reference.dataset  # pylint: disable=no-member

        vendor_dataset_reference = self.config.third_party_vendor_name
        vendor_dataset_key = (
            vendor_dataset_reference.dataset  # pylint: disable=no-member
        )

        if email_dataset_key != vendor_dataset_key:
            error_log = f"Dynamic Erasure Email Connector with key {self.configuration.key} references different datasets for email and vendor fields. Erasure emails not sent."
            logger.error(error_log)
            self.error_all_privacy_requests(
                db,
                privacy_requests,
                "Connector configuration references different datasets for email and vendor fields.",
            )
            raise DynamicErasureEmailConnectorException(error_log)

        collection_name, email_field_name = (
            self.get_collection_and_field_from_reference(
                db, privacy_requests, self.config.recipient_email_address
            )
        )

        vendor_collection_name, vendor_field_name = (
            self.get_collection_and_field_from_reference(
                db, privacy_requests, self.config.third_party_vendor_name
            )
        )

        if collection_name != vendor_collection_name:
            error_log = f"Dynamic Erasure Email Connector with key {self.configuration.key} references different collections for email and vendor fields. Erasure emails not sent."
            logger.error(error_log)
            self.error_all_privacy_requests(
                db,
                privacy_requests,
                "Connector configuration references different collections for email and vendor fields.",
            )
            raise DynamicErasureEmailConnectorException(error_log)

        # We get the DatasetConfig instance using the dataset key provided in the configuration
        dataset_config: Optional[DatasetConfig] = DatasetConfig.get_by(
            db, field="fides_key", value=email_dataset_key
        )

        if not dataset_config:
            error_log = f"DatasetConfig with key {email_dataset_key} not found. Failed to send dynamic erasure emails for connector: {self.configuration.key}."
            logger.error(error_log)
            self.error_all_privacy_requests(
                db,
                privacy_requests,
                f"Connector configuration references DatasetConfig with key {email_dataset_key}, but no such DatasetConfig was found.",
            )

            raise DynamicErasureEmailConnectorException(error_log)

        # Get the graph dataset from the dataset config
        graph_dataset: GraphDataset = dataset_config.get_graph()

        # Get the corresponding ConnectionConfig instance
        connection_config = (
            db.query(ConnectionConfig)
            .filter(ConnectionConfig.id == dataset_config.connection_config_id)
            .first()
        )

        # Instantiate the ConnectionConfig's connector so that we can execute the query
        # to retrieve the email addresses based on the custom request fields
        connector = get_connector(connection_config)  # type: ignore

        # Search for the collection in the dataset that contains the custom request fields
        collections = [
            c for c in graph_dataset.collections if c.name == collection_name
        ]

        if not collections:
            error_log = f"Collection with name {collection_name} not found in dataset {graph_dataset.name}. Failed to send dynamic erasure emails for connector: {self.configuration.key}."
            logger.error(error_log)
            self.error_all_privacy_requests(
                db,
                privacy_requests,
                f"Connector configuration references invalid collection {collection_name} for dataset {graph_dataset.name}",
            )
            raise DynamicErasureEmailConnectorException(error_log)

        # there should only be one collection with that name
        collection = collections[0]

        # We build a dictionary to map the name of each custom request field in the DSR
        # to the name of the corresponding field in the collection.
        # e.g for a field with the following structure:
        #  - name: site_id
        #    fides_meta:
        #      custom_request_field: tenant_id
        # We add the { "tenant_id": "site_id" } key-value pair to the dictionary
        # In order to do this, we basically "flip" the collection.custom_request_fields
        # dictionary, interchanging the keys and values.
        dsr_field_to_collection_field: Dict[str, str] = {
            dsr_field: collection_field.string_path
            for collection_field, dsr_field in collection.custom_request_fields().items()
        }

        # Build the CollectionAddress using the dataset name and the collection name
        collection_address = CollectionAddress(graph_dataset.name, collection_name)

        collection_data = json.loads(
            # Serializes with duck-typing behavior, no longer the default in Pydantic v2
            # Needed for serializing nested collection fields
            collection.model_dump_json(serialize_as_any=True)
        )

        return ProcessedConfig(
            graph_dataset,
            connector,
            collection_address,
            collection_data,
            email_field_name,
            vendor_field_name,
            dsr_field_to_collection_field,
        )

    def get_email_and_vendor_from_custom_request_fields(
        self,
        db: Session,
        processed_config: ProcessedConfig,
        privacy_request: PrivacyRequest,
    ) -> Optional[Tuple[str, str]]:
        """
        Gets the custom request fields from the privacy request and uses them to perform a
        standalone query on the connector, in order to retrieve the value of the email address
        we need to send the erasure email to (specified in field_name).

        Returns a tuple of the form (email, vendor_name), or None if an error occurred or
        some data was not valid.
        """
        # The following code is a bit hacky, but we need to instantiate a RequestTask
        # and an Execution node in order to execute a standalone retrieval query on the connector.
        # The only fields we really care about in the RequestTask are those related to the collection
        # and to the dataset itself.
        # TODO: Once custom request fields can be processed as part of the DSR graph, this logic
        # will be simplified a lot, since the query for the email address will be performed at the
        # Node level as part of the graph traversal and execution. Here we will just need to retrieve
        # it from its corresponding (real) RequestTask and use it to send the email.

        collection_address = processed_config.collection_address

        # Create an in-memory request task, needed for the execution node
        request_task = RequestTask(
            privacy_request_id=privacy_request.id,
            collection_address=collection_address.value,
            dataset_name=collection_address.dataset,
            collection=processed_config.collection_data,
            traversal_details=TraversalDetails.create_empty_traversal(
                processed_config.graph_dataset.connection_key
            ).model_dump(mode="json"),
        )
        execution_node = ExecutionNode(request_task)

        # Get the custom request fields from the privacy request.
        # Returns something like {"tenant_id": {"value": "1234", "label": "Tenant Id"}}
        custom_request_fields = (
            privacy_request.get_persisted_custom_privacy_request_fields()
        )

        # We use the dsr_field_to_collection_field dictionary to build the query filters,
        # where the keys are the names of the custom request fields in the collection (i.e
        # the column names) and the values are the values of the custom request fields in the
        # privacy request.
        # e.g if dsr_field_to_collection_field is {"tenant_id": "site_id"} and
        # custom_request_fields is {"tenant_id": {"value": "1234", "label": "Tenant Id"}},
        # then query_filters will be {"site_id": ["1234"]}
        query_filters = {
            processed_config.dsr_field_to_collection_field[dsr_field]: [
                dsr_field_data["value"]
            ]
            for dsr_field, dsr_field_data in custom_request_fields.items()
        }

        try:
            # We execute a standalone retrieval query on the connector to get the email address
            # based on the custom request fields on the privacy request and the field_name
            # provided as part of the connection config.
            # Following the previous example, this query would be something like
            # SELECT email_address FROM site_info WHERE site_id = '1234'
            retrieved_data = (
                processed_config.connector.execute_standalone_retrieval_query(
                    node=execution_node,
                    fields=[
                        processed_config.email_field,
                        processed_config.vendor_field,
                    ],
                    filters=query_filters,
                )
            )
        except OperationalError:
            error_log = f"An OperationalError occurred when retrieving email from custom request fields for connector {self.configuration.key}. Erasure email for privacy request with id {privacy_request.id} not sent."
            logger.error(error_log)
            self.error_privacy_request(db, privacy_request, error_log)
            return None
        except InternalError:
            error_log = f"An InternalError occurred when retrieving email from custom request fields for connector {self.configuration.key}. Erasure email for privacy request with id {privacy_request.id} not sent."
            logger.error(error_log)
            self.error_privacy_request(db, privacy_request, error_log)
            return None
        except Exception:
            error_log = f"An error occurred when retrieving email from custom request fields for connector {self.configuration.key}. Erasure email for privacy request with id {privacy_request.id} not sent."
            logger.error(error_log)
            self.error_privacy_request(db, privacy_request, error_log)
            return None

        if not retrieved_data:
            logger.error(
                "Custom request field lookup yielded no results. Erasure email for privacy request with id {} not sent.",
                privacy_request.id,
            )
            self.error_privacy_request(
                db,
                privacy_request,
                "Custom request field lookup produced no results: no email address matching custom request fields was found.",
            )
            return None

        if len(retrieved_data) > 1:
            logger.error(
                "Custom request field lookup yielded multiple results. Erasure email for privacy request with id {} not sent.",
                privacy_request.id,
            )
            self.error_privacy_request(
                db,
                privacy_request,
                "Custom request field lookup produced multiple results: multiple email addresses returned for provided custom fields.",
            )
            return None

        email = retrieved_data[0][processed_config.email_field]
        vendor = retrieved_data[0][processed_config.vendor_field]

        return email, vendor

    def test_connection(self) -> Optional[ConnectionTestStatus]:
        """
        Sends an email to the "test_email" configured, just to establish that the email workflow is working.
        """
        logger.info("Starting test connection to {}", self.configuration.key)

        db = Session.object_session(self.configuration)

        try:
            if not self.config.test_email_address:
                raise MessageDispatchException(
                    f"Cannot test connection. No test email defined for {self.configuration.key}"
                )
            # synchronous for now since failure to send is considered a connection test failure
            send_single_erasure_email(
                db=db,
                subject_email=self.config.test_email_address,
                subject_name="Test Vendor",  # Needs to be a string since vendor depends on custom request field values
                batch_identities=list(self.identities_for_test_email.values()),
                test_mode=True,
            )
        except MessageDispatchException as exc:
            logger.info("Email connector test failed with exception {}", exc)
            return ConnectionTestStatus.failed
        return ConnectionTestStatus.succeeded



"""preferences and notices_served received at

Revision ID: efddde14da21
Revises: 4b2eade4353c
Create Date: 2024-05-29 16:17:23.357583

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "efddde14da21"
down_revision = "4b2eade4353c"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "privacypreferencehistory",
        sa.Column("received_at", sa.DateTime(timezone=True), nullable=True),
    )
    op.add_column(
        "servednoticehistory",
        sa.Column("received_at", sa.DateTime(timezone=True), nullable=True),
    )


def downgrade():
    op.drop_column("servednoticehistory", "received_at")
    op.drop_column("privacypreferencehistory", "received_at")



from __future__ import annotations

from typing import Any, Dict, List, Optional, Type

from pydantic import ConfigDict
from sqlalchemy import Boolean, Column, String
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import RelationshipProperty, Session, relationship

from fides.api.db.base_class import Base
from fides.api.models.property import Property
from fides.api.schemas.messaging.messaging import MessagingActionType

# Provides default values for initializing the database or replacing deleted values for messaging templates.
# Note: There are additional MessagingActionTypes that are internally used but are not exposed for user customization.
DEFAULT_MESSAGING_TEMPLATES: Dict[str, Any] = {
    MessagingActionType.SUBJECT_IDENTITY_VERIFICATION.value: {
        "label": "Subject identity verification",
        "content": {
            "subject": "Your one-time code is __CODE__",
            "body": "Your privacy request verification code is __CODE__. Please return to the Privacy Center and enter the code to continue. This code will expire in __MINUTES__ minutes.",
        },
    },
    MessagingActionType.PRIVACY_REQUEST_RECEIPT.value: {
        "label": "Privacy request received",
        "content": {
            "subject": "Your privacy request has been received",
            "body": "Your privacy request has been received. We will get back to you shortly.",
        },
    },
    MessagingActionType.PRIVACY_REQUEST_REVIEW_APPROVE.value: {
        "label": "Privacy request approved",
        "content": {
            "subject": "Your privacy request has been approved",
            "body": "Your privacy request has been approved and is currently processing.",
        },
    },
    MessagingActionType.PRIVACY_REQUEST_REVIEW_DENY.value: {
        "label": "Privacy request denied",
        "content": {
            "subject": "Your privacy request has been denied",
            "body": "Your privacy request has been denied. __DENIAL_REASON__.",
        },
    },
    MessagingActionType.PRIVACY_REQUEST_COMPLETE_ACCESS.value: {
        "label": "Access request completed",
        "content": {
            "subject": "Your data is ready to be downloaded",
            "body": "Your access request has been completed and can be downloaded at __DOWNLOAD_LINK__. For security purposes, this secret link will expire in __DAYS__ days.",
        },
    },
    MessagingActionType.PRIVACY_REQUEST_COMPLETE_DELETION.value: {
        "label": "Erasure request completed",
        "content": {
            "subject": "Your data has been deleted",
            "body": "Your erasure request has been completed.",
        },
    },
}


def _link_properties_to_messaging_template(
    db: Session,
    properties: List[Dict[str, Any]],
    messaging_template: MessagingTemplate,
) -> Optional[List[Property]]:
    """
    Link supplied properties to MessagingTemplate and unlink any properties not supplied.
    """
    new_properties = []
    if len(properties) > 0:
        new_properties = (
            db.query(Property)
            .filter(Property.id.in_([prop["id"] for prop in properties]))
            .all()
        )
    messaging_template.properties = new_properties
    messaging_template.save(db)
    return messaging_template.properties


class MessagingTemplate(Base):
    @declared_attr
    def __tablename__(self) -> str:
        return "messaging_template"

    type = Column(String, index=True, nullable=False)
    content = Column(MutableDict.as_mutable(JSONB), nullable=False)
    properties: RelationshipProperty[List[Property]] = relationship(
        "Property",
        secondary="messaging_template_to_property",
        back_populates="messaging_templates",
        lazy="selectin",
    )
    is_enabled = Column(Boolean, default=False, nullable=False)
    model_config = ConfigDict(from_attributes=True)

    @classmethod
    def create(
        cls: Type[MessagingTemplate],
        db: Session,
        *,
        data: dict[str, Any],
        check_name: bool = True,
    ) -> MessagingTemplate:
        """
        Creates a Messaging Template, allows linking properties
        """
        properties = data.pop("properties", [])
        messaging_template: MessagingTemplate = super().create(
            db=db, data=data, check_name=check_name
        )
        # Link Properties to this Messaging Template via the MessagingTemplateToProperty table
        _link_properties_to_messaging_template(db, properties, messaging_template)

        return messaging_template

    def update(self, db: Session, *, data: dict[str, Any]) -> MessagingTemplate:
        """
        Updates a Messaging Template, allows linking properties
        """
        properties = data.pop("properties", [])
        super().update(db=db, data=data)

        # Link Properties to this Messaging Template via the MessagingTemplateToProperty table
        _link_properties_to_messaging_template(db, properties, self)
        return self



"""adds DRP action to policy

Revision ID: 5078badb90b9
Revises: c98da12d76f8
Create Date: 2022-05-04 17:22:46.500067

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
from sqlalchemy.dialects import postgresql

revision = "5078badb90b9"
down_revision = "c98da12d76f8"
branch_labels = None
depends_on = None


def upgrade():
    drpaction = postgresql.ENUM(
        "access",
        "deletion",
        "sale_opt_out",
        "sale_opt_in",
        "access_categories",
        "access_specific",
        name="drpaction",
        create_type=False,
    )
    drpaction.create(op.get_bind())
    op.add_column("policy", sa.Column("drp_action", drpaction, nullable=True))
    op.create_index(op.f("ix_policy_drp_action"), "policy", ["drp_action"], unique=True)


def downgrade():
    op.drop_index(op.f("ix_policy_drp_action"), table_name="policy")
    op.drop_column("policy", "drp_action")
    op.execute("DROP TYPE drpaction;")



from enum import Enum
from typing import Any, Dict, List

from pydantic import ValidationError

from fides.api.common_exceptions import NoSuchStrategyException
from fides.api.common_exceptions import ValidationError as FidesopsValidationError
from fides.api.schemas.saas.strategy_configuration import StrategyConfiguration
from fides.api.service.async_dsr.async_dsr_strategy import AsyncDSRStrategy
from fides.api.service.async_dsr.async_dsr_strategy_callback import (
    CallbackAsyncDSRStrategy,
)
from fides.api.service.async_dsr.async_dsr_strategy_polling import (
    PollingAsyncDSRStrategy,
)


class SupportedAsyncDSRStrategies(Enum):
    """
    The supported async DSR strategies for handling asynchronous data subject requests.
    """

    callback = CallbackAsyncDSRStrategy
    polling = PollingAsyncDSRStrategy

    @classmethod
    def __contains__(cls, item: str) -> bool:
        try:
            cls[item]
        except KeyError:
            return False

        return True


def get_strategy(
    strategy_name: str,
    configuration: Dict[str, Any],
) -> AsyncDSRStrategy:
    """
    Returns the strategy given the name and configuration.
    Raises NoSuchStrategyException if the strategy does not exist
    """
    if not SupportedAsyncDSRStrategies.__contains__(strategy_name):
        valid_strategies = ", ".join(get_strategy_names())
        raise NoSuchStrategyException(
            f"Strategy '{strategy_name}' does not exist. Valid strategies are [{valid_strategies}]"
        )
    strategy = SupportedAsyncDSRStrategies[strategy_name].value
    if strategy.configuration_model is None:
        return strategy()
    try:
        strategy_config: StrategyConfiguration = strategy.configuration_model(
            **configuration
        )
        return strategy(configuration=strategy_config)
    except ValidationError as e:
        raise FidesopsValidationError(message=str(e))


def get_strategy_names() -> List[str]:
    """Returns all supported async DSR strategies"""
    return [s.name for s in SupportedAsyncDSRStrategies]



"""add location to privacy request

Revision ID: b1a2c3d4e5f6
Revises: 2f3c1a2d6b10
Create Date: 2025-08-06 18:15:00.000000

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "b1a2c3d4e5f6"
down_revision = "90502bcda282"
branch_labels = None
depends_on = None


def upgrade():
    # Add location column to privacyrequest table
    op.add_column("privacyrequest", sa.Column("location", sa.String(), nullable=True))


def downgrade():
    # Remove location column from privacyrequest table
    op.drop_column("privacyrequest", "location")



"""Contains the user command group for the fides CLI."""

import rich_click as click
from pydantic import EmailStr

from fides.cli.options import (
    email_address_argument,
    first_name_option,
    last_name_option,
    password_argument,
    password_option,
    username_argument,
    username_option,
)
from fides.cli.utils import with_server_health_check
from fides.core.user import create_command, get_permissions_command, login_command


@click.group(name="user")
@click.pass_context
def user(ctx: click.Context) -> None:
    """
    Click command group for interacting with user-related functionality.
    """


@user.command()
@click.pass_context
@username_argument
@password_argument
@email_address_argument
@first_name_option
@last_name_option
@with_server_health_check
def create(
    ctx: click.Context,
    username: str,
    password: str,
    email_address: EmailStr,
    first_name: str,
    last_name: str,
) -> None:
    """
    Use the credentials file to create a new user. Gives full permissions to the new user.
    """
    config = ctx.obj["CONFIG"]
    server_url = config.cli.server_url
    create_command(
        username=username,
        password=password,
        email_address=email_address,
        first_name=first_name,
        last_name=last_name,
        server_url=server_url,
    )


@user.command()
@click.pass_context
@username_option
@password_option
@with_server_health_check
def login(ctx: click.Context, username: str, password: str) -> None:
    """
    Authenticate with the webserver and generate a user access token.
    Then store those credentials in a credentials file.
    """
    config = ctx.obj["CONFIG"]
    server_url = config.cli.server_url
    login_command(username=username, password=password, server_url=server_url)


@user.command(name="permissions")
@click.pass_context
@with_server_health_check
def get_permissions(ctx: click.Context) -> None:
    """
    List the directly-assigned scopes and roles available to the current user.
    """
    config = ctx.obj["CONFIG"]
    server_url = config.cli.server_url
    get_permissions_command(server_url=server_url)



from typing import Any, Callable, Dict, List, Optional, Set, Tuple

from fideslang.validation import FidesKey
from loguru import logger

from fides.api.graph.config import (
    TERMINATOR_ADDRESS,
    Collection,
    CollectionAddress,
    Field,
    FieldAddress,
    FieldPath,
)
from fides.api.graph.graph import Edge
from fides.api.models.privacy_request import RequestTask, TraversalDetails
from fides.api.util.collection_util import partition
from fides.api.util.logger_context_utils import Contextualizable, LoggerContextKeys

COLLECTION_FIELD_PATH_MAP = Dict[CollectionAddress, List[Tuple[FieldPath, FieldPath]]]


class ExecutionNode(Contextualizable):  # pylint: disable=too-many-instance-attributes
    """Node for *executing* a task. This node only has knowledge of itself and its incoming and outgoing edges

    After we build the graph, we save details to RequestTasks in the database that are hydrated here to execute an individual
    node without rebuilding the graph with traversal.traverse
    """

    def __init__(self, request_task: RequestTask):
        assert request_task.collection  # For mypy
        self.collection: Collection = Collection.parse_from_request_task(
            request_task.collection
        )
        self.address: CollectionAddress = CollectionAddress.from_string(
            request_task.collection_address
        )
        traversal_details = TraversalDetails.model_validate(
            request_task.traversal_details or {}
        )

        self.incoming_edges: Set[Edge] = {
            Edge(FieldAddress.from_string(edge[0]), FieldAddress.from_string(edge[1]))
            for edge in traversal_details.incoming_edges
        }
        self.outgoing_edges: Set[Edge] = {
            Edge(FieldAddress.from_string(edge[0]), FieldAddress.from_string(edge[1]))
            for edge in traversal_details.outgoing_edges
        }
        self.connection_key: FidesKey = FidesKey(
            traversal_details.dataset_connection_key
        )

        self.incoming_edges_by_collection: Dict[CollectionAddress, List[Edge]] = (
            partition(self.incoming_edges, lambda e: e.f1.collection_address())
        )

        # Input should be passed into accessing data in this order
        self.input_keys: List[CollectionAddress] = [
            CollectionAddress.from_string(input_key)
            for input_key in traversal_details.input_keys
        ]
        self.grouped_fields = self.collection.grouped_inputs

    @property
    def query_field_paths(self) -> Set[FieldPath]:
        """
        All of the possible field paths that we can query for possible filter values.
        These are field paths that are the ends of incoming edges.
        """
        return {edge.f2.field_path for edge in self.incoming_edges}

    @property
    def dependent_identity_fields(self) -> bool:
        """If the current collection needs inputs from other collections, in addition to its seed data."""
        for field in self.grouped_fields:
            if self.collection.field(FieldPath(field)).identity:  # type: ignore
                return True
        return False

    def get_log_context(self) -> Dict[LoggerContextKeys, Any]:
        return {LoggerContextKeys.collection: self.collection.name}

    def build_incoming_field_path_maps(
        self, group_dependent_fields: bool = False
    ) -> Tuple[COLLECTION_FIELD_PATH_MAP, COLLECTION_FIELD_PATH_MAP]:
        """
        For each collection connected to the current collection, return a list of tuples
        mapping the foreign field to the local field.  This is used to process data from incoming collections
        into the current collection.

        :param group_dependent_fields: Whether we should split the incoming fields into two groups: one whose
        fields are completely independent of one another, and the other whose incoming data needs to stay linked together.
        If False, all fields are returned in the first tuple, and the second tuple just maps collections to an empty list.

        """

        def field_map(keep: Callable) -> COLLECTION_FIELD_PATH_MAP:
            return {
                col_addr: [
                    (edge.f1.field_path, edge.f2.field_path)
                    for edge in edge_list
                    if keep(edge.f2.field_path.string_path)
                ]
                for col_addr, edge_list in self.incoming_edges_by_collection.items()
            }

        if group_dependent_fields:
            return field_map(
                lambda string_path: string_path not in self.grouped_fields
            ), field_map(lambda string_path: string_path in self.grouped_fields)

        return field_map(lambda string_path: True), field_map(lambda string_path: False)

    def typed_filtered_values(self, input_data: Dict[str, List[Any]]) -> Dict[str, Any]:
        """
        Return a filtered list of key/value sets of data items that are both in
        the list of incoming edge fields, and contain data in the input data set.

        The values are cast based on field types, if those types are specified.
        """
        out = {}
        failed_conversions: Dict[str, Dict[str, Any]] = {}

        for key, values in input_data.items():
            path: FieldPath = FieldPath.parse(key)
            field: Optional[Field] = self.collection.field(path)

            if field and path in self.query_field_paths and isinstance(values, list):
                cast_values = [field.cast(v) for v in values]
                filtered = list(filter(lambda x: x is not None, cast_values))
                if filtered:
                    out[key] = filtered
                elif values:  # Had input values but all failed conversion
                    # Track conversion failures for logging
                    failed_conversions[key] = {"field": field, "input_values": values}

        # Log conversion failures if any occurred
        if failed_conversions:
            field_details = []
            for field_name, failure_info in failed_conversions.items():
                field = failure_info["field"]
                values = failure_info["input_values"]

                expected_type = field.data_type() if field else "unknown"
                actual_types = {type(v).__name__ for v in values if v is not None}
                actual_type_str = (
                    ", ".join(sorted(actual_types)) if actual_types else "NoneType"
                )

                field_details.append(
                    f"{field_name} (expected: {expected_type}, got: {actual_type_str})"
                )

            logger.warning(
                "Type conversion failures for {}: {}",
                self.address,
                "; ".join(field_details),
            )

        return out

    def has_outgoing_dependencies(self) -> bool:
        """
        Check if this node has outgoing edges to collections other than the terminal node.

        Returns:
            bool: True if the node has dependencies on other collections, False otherwise
        """

        for edge in self.outgoing_edges:
            # Check if the outgoing edge points to a collection other than the terminal node
            target_collection = edge.f2.collection_address()
            if target_collection != TERMINATOR_ADDRESS:
                return True
        return False



"""
Contains utility functions that set up the application webserver.
"""

# pylint: disable=too-many-branches
from logging import DEBUG
from typing import AsyncGenerator, List

from fastapi import FastAPI
from fastapi.middleware.gzip import GZipMiddleware
from fastapi.routing import APIRoute
from loguru import logger
from redis.exceptions import RedisError, ResponseError
from slowapi.errors import RateLimitExceeded  # type: ignore
from slowapi.extension import _rate_limit_exceeded_handler  # type: ignore
from slowapi.middleware import SlowAPIMiddleware  # type: ignore

import fides
from fides.api.api.deps import (
    get_api_session,
    get_async_autoclose_db_session,
    get_autoclose_db_session,
)
from fides.api.api.v1 import CTL_ROUTER
from fides.api.api.v1.api import api_router
from fides.api.api.v1.endpoints.admin import ADMIN_ROUTER
from fides.api.api.v1.endpoints.generic_overrides import GENERIC_OVERRIDES_ROUTER
from fides.api.api.v1.endpoints.health import HEALTH_ROUTER
from fides.api.api.v1.exception_handlers import ExceptionHandlers
from fides.api.common_exceptions import RedisConnectionError, RedisNotConfigured
from fides.api.db import seed
from fides.api.db.database import configure_db, seed_db
from fides.api.db.seed import create_or_update_parent_user
from fides.api.models.application_config import ApplicationConfig
from fides.api.oauth.system_manager_oauth_util import (
    get_system_fides_key,
    get_system_schema,
    verify_oauth_client_for_system_from_fides_key_cli,
    verify_oauth_client_for_system_from_request_body_cli,
)
from fides.api.oauth.utils import get_root_client, verify_oauth_client_prod
from fides.api.service.connectors.saas.connector_registry_service import (
    update_saas_configs,
)

# pylint: disable=wildcard-import, unused-wildcard-import
from fides.api.service.saas_request.override_implementations import *
from fides.api.util.api_router import APIRouter
from fides.api.util.cache import get_cache
from fides.api.util.consent_util import create_default_tcf_purpose_overrides_on_startup
from fides.api.util.errors import FidesError
from fides.api.util.logger import setup as setup_logging
from fides.api.util.rate_limit import (
    RateLimitIPValidationMiddleware,
    fides_limiter,
    is_rate_limit_enabled,
)
from fides.config import CONFIG
from fides.config.config_proxy import ConfigProxy

VERSION = fides.__version__

# DB_ROUTER holds routers that have direct DB dependencies.
# these routers are initialized _outside_ of inner `api` module
# to avoid cyclical dependency chains.
# see https://github.com/ethyca/fides/issues/3652
DB_ROUTER = APIRouter()
DB_ROUTER.include_router(ADMIN_ROUTER)
DB_ROUTER.include_router(HEALTH_ROUTER)


ROUTERS = [CTL_ROUTER, api_router, DB_ROUTER]
OVERRIDING_ROUTERS = [GENERIC_OVERRIDES_ROUTER]


def create_fides_app(
    lifespan: AsyncGenerator[None, None],
    routers: List = ROUTERS,
    app_version: str = VERSION,
    security_env: str = CONFIG.security.env,
) -> FastAPI:
    """Return a properly configured application."""
    setup_logging(CONFIG)
    logger.bind(api_config=CONFIG.logging.json()).debug(
        "Logger configuration options in use"
    )

    fastapi_app = FastAPI(title="fides", version=app_version, lifespan=lifespan, separate_input_output_schemas=False)  # type: ignore
    fastapi_app.state.limiter = fides_limiter
    # Starlette bug causing this to fail mypy
    fastapi_app.add_exception_handler(RateLimitExceeded, _rate_limit_exceeded_handler)  # type: ignore
    for handler in ExceptionHandlers.get_handlers():
        # Starlette bug causing this to fail mypy
        fastapi_app.add_exception_handler(RedisNotConfigured, handler)  # type: ignore

    if is_rate_limit_enabled:
        # Validate header before SlowAPI processes the request
        fastapi_app.add_middleware(RateLimitIPValidationMiddleware)
        # Required for default rate limiting to work
        fastapi_app.add_middleware(SlowAPIMiddleware)
    else:
        logger.warning(
            "Rate limiting client IPs is disabled because the FIDES__SECURITY__RATE_LIMIT_CLIENT_IP_HEADER env var is not configured."
        )

    fastapi_app.add_middleware(
        GZipMiddleware, minimum_size=1000, compresslevel=5
    )  # minimum_size is in bytes

    for router in routers:
        fastapi_app.include_router(router)

    override_generic_routers(OVERRIDING_ROUTERS, fastapi_app)

    if security_env == "dev":
        # This removes auth requirements for specific endpoints
        fastapi_app.dependency_overrides[verify_oauth_client_prod] = get_root_client
        fastapi_app.dependency_overrides[
            verify_oauth_client_for_system_from_request_body_cli
        ] = get_system_schema
        fastapi_app.dependency_overrides[
            verify_oauth_client_for_system_from_fides_key_cli
        ] = get_system_fides_key
    elif security_env == "prod":
        # This is the most secure, so all security deps are maintained
        pass

    return fastapi_app


def override_generic_routers(
    overriding_routers: List[APIRouter], base_router: FastAPI
) -> None:
    """
    Remove generic routes in favor of their more specific implementations, if available.
    """
    for i, existing_route in reversed(list(enumerate(base_router.routes))):
        if not isinstance(existing_route, APIRoute):
            continue
        for new_router in overriding_routers:
            for new_route in new_router.routes:
                if not isinstance(new_route, APIRoute):  # pragma: no cover
                    continue
                if (
                    existing_route.methods == new_route.methods
                    and existing_route.path == new_route.path
                ):
                    logger.debug(
                        "Removing generic route: {} {}",
                        existing_route.methods,
                        existing_route.path,
                    )
                    del base_router.routes[i]
    for router in overriding_routers:
        base_router.include_router(router)


def log_startup() -> None:
    """Log application startup and other information."""
    logger.info(f"Starting Fides - v{VERSION}")
    logger.info(
        "Startup configuration: reloading = {}, dev_mode = {}",
        CONFIG.hot_reloading,
        CONFIG.dev_mode,
    )
    logger.info("Startup configuration: pii logging = {}", CONFIG.logging.log_pii)

    if CONFIG.logging.level == DEBUG:
        logger.warning(
            "WARNING: log level is DEBUG, so sensitive or personal data may be logged. "
            "Set FIDES__LOGGING__LEVEL to INFO or higher in production."
        )
        CONFIG.log_all_config_values()


async def run_database_startup(app: FastAPI) -> None:
    """
    Perform all relevant database startup activities/configuration for the
    application webserver.
    """

    if not CONFIG.database.sync_database_uri:
        raise FidesError("No database uri provided")

    if CONFIG.database.automigrate:
        try:
            configure_db(CONFIG.database.sync_database_uri)
            if not CONFIG.test_mode:
                with get_autoclose_db_session() as session:
                    seed_db(session)
            if CONFIG.database.load_samples:
                async with get_async_autoclose_db_session() as async_session:
                    await seed.load_samples(async_session)
        except Exception as e:
            logger.error("Error occurred during database configuration: {}", str(e))
    else:
        logger.info("Skipping auto-migration due to 'automigrate' configuration value.")

    try:
        create_or_update_parent_user()
    except Exception as e:
        logger.error("Error creating parent user: {}", str(e))
        raise FidesError(f"Error creating parent user: {str(e)}")

    db = get_api_session()
    logger.info("Loading config settings into database...")
    try:
        ApplicationConfig.update_config_set(db, CONFIG)
    except Exception as e:
        logger.error("Error occurred writing config settings to database: {}", str(e))
        raise FidesError(
            f"Error occurred writing config settings to database: {str(e)}"
        )
    finally:
        db.close()

    logger.info("Loading CORS domains from ConfigProxy...")
    try:
        ConfigProxy(db).load_current_cors_domains_into_middleware(app)
    except Exception as e:
        logger.error("Error occurred while loading CORS domains: {}", str(e))
        raise FidesError(f"Error occurred while loading CORS domains: {str(e)}")
    finally:
        db.close()

    logger.info("Validating SaaS connector templates...")
    try:
        update_saas_configs(db)
        logger.info("Finished loading SaaS templates")
    except Exception as e:
        logger.exception(
            "Error occurred during SaaS connector template validation: {}",
            str(e),
        )
        return
    finally:
        db.close()

    if not CONFIG.test_mode:
        # Avoiding loading consent out-of-the-box resources to avoid interfering with unit tests
        load_tcf_purpose_overrides()

    db.close()


def check_redis() -> None:
    """Check that Redis is healthy."""
    logger.info("Running Cache connection test...")
    try:
        get_cache()
    except (RedisConnectionError, RedisError, ResponseError) as e:
        logger.error("Connection to cache failed: {}", str(e))
        return

    logger.debug("Connection to cache succeeded")


def load_tcf_purpose_overrides() -> None:
    """Load default tcf purpose overrides"""
    logger.info("Loading default TCF Purpose Overrides")
    try:
        db = get_api_session()
        create_default_tcf_purpose_overrides_on_startup(db)
    except Exception as e:
        logger.error("Skipping loading TCF Purpose Overrides: {}", str(e))
    finally:
        db.close()



"""Schema for external storage metadata."""

from typing import Optional

from pydantic import Field

from fides.api.schemas.base_class import FidesSchema
from fides.api.schemas.storage.storage import StorageType


class ExternalStorageMetadata(FidesSchema):
    """Metadata for externally stored encrypted data."""

    storage_type: StorageType
    file_key: str = Field(description="Path/key of the file in external storage")
    filesize: int = Field(description="Size of the stored file in bytes", ge=0)
    storage_key: Optional[str] = Field(
        default=None, description="Storage configuration key used"
    )

    class Config:
        use_enum_values = True



from .consent import Consent, ConsentRequest
from .execution_log import (
    COMPLETED_EXECUTION_LOG_STATUSES,
    EXITED_EXECUTION_LOG_STATUSES,
    ExecutionLog,
    can_run_checkpoint,
)
from .privacy_request import (
    CustomPrivacyRequestField,
    PrivacyRequest,
    PrivacyRequestError,
    PrivacyRequestNotifications,
)
from .provided_identity import ProvidedIdentity, ProvidedIdentityType
from .request_task import RequestTask, TraversalDetails
from .webhook import (
    CallbackType,
    SecondPartyRequestFormat,
    generate_request_callback_pre_approval_jwe,
    generate_request_callback_resume_jwe,
    generate_request_task_callback_jwe,
)

__all__ = [
    "CallbackType",
    "Consent",
    "ConsentRequest",
    "COMPLETED_EXECUTION_LOG_STATUSES",
    "EXITED_EXECUTION_LOG_STATUSES",
    "CustomPrivacyRequestField",
    "ExecutionLog",
    "PrivacyRequest",
    "PrivacyRequestError",
    "PrivacyRequestNotifications",
    "ProvidedIdentity",
    "ProvidedIdentityType",
    "RequestTask",
    "SecondPartyRequestFormat",
    "TraversalDetails",
    "can_run_checkpoint",
    "generate_request_callback_pre_approval_jwe",
    "generate_request_callback_resume_jwe",
    "generate_request_task_callback_jwe",
]



from fides.api.service.connectors.query_configs.query_config import (
    QueryStringWithoutTuplesOverrideQueryConfig,
)


class MicrosoftSQLServerQueryConfig(QueryStringWithoutTuplesOverrideQueryConfig):
    """
    Generates SQL valid for SQLServer.
    """



from typing import Optional, cast

from fides.api.task.conditional_dependencies.schemas import (
    ConditionEvaluationResult,
    EvaluationResult,
    GroupEvaluationResult,
)

MAX_DEPTH = 100


def format_evaluation_success_message(
    evaluation_result: Optional[EvaluationResult],
) -> str:
    """Format a detailed message about which conditions succeeded

    Args:
        evaluation_result: The evaluation result to create a string from

    Returns:
        A string describing the evaluation result

    """
    if not evaluation_result:
        return "No conditional dependencies to evaluate."

    return _format_evaluation_message(evaluation_result, success=True, depth=0)


def format_evaluation_failure_message(
    evaluation_result: Optional[EvaluationResult],
) -> str:
    """Format a detailed message about which conditions failed

    Args:
        evaluation_result: The evaluation result to create a string from

    Returns:
        A string describing the evaluation result
    """
    if not evaluation_result:
        return "No conditional dependencies to evaluate."

    return _format_evaluation_message(evaluation_result, success=False, depth=0)


def _format_leaf_condition(evaluation_result: ConditionEvaluationResult) -> str:
    """Format a single leaf condition into a readable string

    Args:
        evaluation_result: The conditionevaluation result to create a string from

    Returns:
        A string describing the evaluation result
    """
    condition_desc = f"{evaluation_result.field_address} {evaluation_result.operator}"
    if evaluation_result.expected_value is not None:
        condition_desc += f" {evaluation_result.expected_value}"
    return condition_desc


def _format_condition_list(
    results: list[EvaluationResult], success: bool, depth: int
) -> list[str]:
    """Format a list of conditions (either leaf or group) into readable strings

    Args:
        results: The list of conditions to format
        success: Whether the conditions were successful
        depth: The depth of the conditions

    Returns:
        A list of strings describing the conditions
    """
    condition_descriptions = []
    for sub_result in results:
        if _is_leaf_condition(sub_result):
            # Cast to the specific type after verification
            leaf_result = cast(ConditionEvaluationResult, sub_result)
            condition_descriptions.append(_format_leaf_condition(leaf_result))
        elif _is_group_condition(sub_result):
            # Cast to the specific type after verification
            group_result = cast(GroupEvaluationResult, sub_result)
            condition_descriptions.append(
                _format_evaluation_message(group_result, success, depth + 1)
            )
    return condition_descriptions


def _format_evaluation_message(
    evaluation_result: EvaluationResult, success: bool, depth: int = 0
) -> str:
    """Format evaluation results into a readable message

    Args:
        evaluation_result: The evaluation result to format
        success: Whether the conditions were successful
        depth: The depth of the conditions

    Returns:
        A string describing the evaluation result
    """
    # Prevent infinite recursion
    if depth > MAX_DEPTH:
        return "Condition evaluation too deeply nested"

    # Try to format as group condition first
    if _is_group_condition(evaluation_result):
        # Cast to the specific type after verification
        group_result = cast(GroupEvaluationResult, evaluation_result)
        return _format_group_condition(group_result, success, depth)

    # Try to format as leaf condition
    if _is_leaf_condition(evaluation_result):
        # Cast to the specific type after verification
        leaf_result = cast(ConditionEvaluationResult, evaluation_result)
        return _format_leaf_condition_message(leaf_result, success)

    # Unknown condition type
    return "Evaluation result details unavailable"


def _is_group_condition(evaluation_result: EvaluationResult) -> bool:
    """Check if evaluation_result is a group condition by checking for group-specific attributes

    Args:
        evaluation_result: The evaluation result to check

    Returns:
        True if the evaluation result is a group condition, False otherwise
    """
    # Check for attributes that are unique to GroupEvaluationResult
    return hasattr(evaluation_result, "logical_operator") and hasattr(
        evaluation_result, "condition_results"
    )


def _is_leaf_condition(evaluation_result: EvaluationResult) -> bool:
    """Check if evaluation_result is a leaf condition by checking for leaf-specific attributes

    Args:
        evaluation_result: The evaluation result to check

    Returns:
        True if the evaluation result is a leaf condition, False otherwise
    """
    # Check for attributes that are unique to ConditionEvaluationResult
    return hasattr(evaluation_result, "field_address") and hasattr(
        evaluation_result, "operator"
    )


def _format_group_condition(
    evaluation_result: GroupEvaluationResult, success: bool, depth: int
) -> str:
    """Format a group condition evaluation result

    Args:
        evaluation_result: The group evaluation result to format
        success: Whether the conditions were successful
        depth: The depth of the conditions

    Returns:
        A string describing the evaluation result
    """
    logical_operator = evaluation_result.logical_operator
    results = evaluation_result.condition_results
    condition_descriptions = _format_condition_list(results, success, depth)

    if success:
        return f"All conditions in {logical_operator.upper()} group were met: {'; '.join(condition_descriptions)}"

    if condition_descriptions:
        return f"Failed conditions in {logical_operator.upper()} group: {'; '.join(condition_descriptions)}"

    return f"Group condition with {logical_operator.upper()} operator failed"


def _format_leaf_condition_message(
    evaluation_result: ConditionEvaluationResult, success: bool
) -> str:
    """Format a leaf condition evaluation result

    Args:
        evaluation_result: The leaf evaluation result to format
        success: Whether the conditions were successful

    Returns:
        A string describing the evaluation result
    """
    condition_desc = _format_leaf_condition(evaluation_result)

    if success:
        return f"Condition '{condition_desc}' was met"

    return f"Condition '{condition_desc}' was not met"



"""Module that adds functionality for resource filtering."""

from typing import Callable, List, Optional

from fideslang.models import Organization, System

IGNORE_RESOURCE_ARN_FILTER = "ignore_resource_arn"
AWS_RESOURCE_FILTERS: List[str] = [IGNORE_RESOURCE_ARN_FILTER]


def get_system_arn(system: System) -> Optional[str]:
    """
    Returns resource arn for System if it is defined
    """
    system_arn = (
        system.fidesctl_meta.resource_id
        if system.fidesctl_meta and system.fidesctl_meta.resource_id
        else None
    )
    return system_arn


def is_arn_filter_match(arn: str, filter_arn: str) -> bool:
    """
    Returns True if the given arn matches the filter pattern. In order
    for an arn to be a match, it must match each field separated by a
    colon(:). If the filter pattern contains an empty field, then it is
    treated as wildcard for that field.

    Examples:
    arn=abc:def:ghi   filter_arn=abc:def:ghi   #exact match returns True
    arn=abc:def:ghi   filter_arn=abc::ghi      #wildcarded match returns True
    """
    arn_split = arn.split(":")
    filter_arn_split = filter_arn.split(":")

    # arns of different field lenth are treated as mismatch
    if len(arn_split) != len(filter_arn_split):
        return False

    for i in range(0, len(arn_split), 1):
        # filter pattern with value defined can be considered for mismatch
        if filter_arn_split[i] and filter_arn_split[i] != arn_split[i]:
            return False
    return True


def ignore_resource_arn(systems: List[System], filter_value: str) -> List[System]:
    """
    Given a list of systems and a filter pattern, returns elements which
    are not excluded by the arn matchin filter
    """
    filtered_systems = []
    for system in systems:
        system_arn = get_system_arn(system=system)
        if not system_arn or not is_arn_filter_match(
            arn=system_arn, filter_arn=filter_value
        ):
            filtered_systems.append(system)
    return filtered_systems


def get_aws_system_filter_function(
    filter_type: str,
) -> Callable[[List[System], str], List[System]]:
    """
    Returns filter function for a given filter type. A filter function
    takes an input System list and returns a new list that excludes filter
    matches.
    """
    system_filter_functions = {
        IGNORE_RESOURCE_ARN_FILTER: ignore_resource_arn,
    }
    filter_function = system_filter_functions[filter_type]
    return filter_function


def filter_aws_systems(
    systems: List[System], organization: Organization
) -> List[System]:
    """
    Given an organization with resource filters, passes a list
    of Systems through each filter. Filter ordering is not important
    as filters can only exclude Systems. Filter type is used to deremine
    the filter function.
    """
    resource_filters = (
        organization.fidesctl_meta.resource_filters
        if organization.fidesctl_meta and organization.fidesctl_meta.resource_filters
        else []
    )
    filtered_systems = systems
    for resource_filter in resource_filters:
        filter_type = resource_filter.type
        if filter_type in AWS_RESOURCE_FILTERS:
            filter_function = get_aws_system_filter_function(resource_filter.type)
            filtered_systems = filter_function(filtered_systems, resource_filter.value)
    return filtered_systems



from fides.api.service.async_dsr.async_dsr_strategy import AsyncDSRStrategy


class CallbackAsyncDSRStrategy(AsyncDSRStrategy):
    """
    Strategy for callback async DSR requests.

    This strategy has no methods since the callback is handled by the callback endpoint
    we still require the strategy to be defined to notify that the request is awaiting async processing
    and follows the AsyncDSRStrategy interface.
    """

    name = "callback"

    def __init__(self) -> None:
        pass
